from ffxivcalc.Jobs.Base_Spell import buff, empty, buffPercentHistory
from ffxivcalc.Jobs.Melee.Melee_Spell import ReaperSpell
Lock = 0

#Requirement

def SoulsowRequirement(Player, Spell):
    # This requirement will only serve to make the casting instant if the fight is not started
    if not Player.CurrentFight.FightStart:
        Spell.CastTime = 0
    return True, -1

def HarvestMoonRequirement(Player, Spell):
    return Player.Soulsow, -1

def SoulSliceRequirement(Player, Spell):
    return Player.SoulSliceStack > 0, Player.SoulSliceCD

def ArcaneCircleRequirement(Player, Spell):
    return Player.ArcaneCircleCD <= 0, Player.ArcaneCircleCD

def GluttonyRequirement(Player, Spell):
    return Player.GluttonyCD <= 0 and Player.SoulGauge >= 50, Player.GluttonyCD

def BloodStalkRequirement(Player, Spell):
    return Player.SoulGauge >= 50, -1

def GibbetRequirement(Player, Spell):
    return Player.SoulReaverStack > 0 and Player.AvatarTimer == 0, -1

def PlentifulHarvestRequirement(Player, Spell):

    #Will modify Spell's potency depending on ImmortalStack, will add 40 potency for each stack, for a max of 1000
    if Player.CurrentFight.MaxPotencyPlentifulHarvest:           # If wants max potency regardless team
        Spell.Potency = 1000
    else:
        Spell.Potency += (Player.ImmortalSacrificeStack - 1) * 40 #Removing minimal requirement, which is 1 stack

    return Player.ImmortalSacrificeStack > 0 and Player.BloodsownTimer == 0, Player.BloodsownTimer

def EnshroudRequirement(Player, Spell):
    return Player.EnshroudCD <= 0 and Player.ShroudGauge >= 50, Player.EnshroudCD


def VoidReapingRequirement(Player, Spell):
    return Player.AvatarTimer > 0 and Player.LemureGauge > 0, -1

def LemureSliceRequirement(Player, Spell):
    return Player.VoidShroudGauge > 1, -1

def CommunioRequirement(Player, Spell):
    return Player.LemureGauge > 0, -1

def UnveiledGibbetRequirement(Player, Spell):
    return Player.EnhancedGibbet, -1

def UnveiledGallowsRequirement(Player, Spell):
    return Player.EnhancedGallows, -1

def HellIngressRequirement(Player, Spell):
    return Player.HellIngressCD <= 0, Player.HellIngressCD

def ArcaneCrestRequirement(Player, Spell):
    return Player.ArcaneCrestCD <= 0, Player.ArcaneCrestCD

#Apply

def ApplyGrimReaping(Player, Enemy):
    Player.LemureGauge -= 1
    Player.VoidShroudGauge = min(5, Player.VoidShroudGauge + 1)

def ApplyGuillotine(Player, Enemy):
    Player.AddShroud(10)

def ApplyArcaneCrest(Player, Enemy):
    Player.ArcaneCrestCD = 30

def ApplySpinningScythe(Player, Enemy):
    Player.AddGauge(10)
    if not (SpinningScytheCombo in Player.EffectList) : Player.EffectList.append(SpinningScytheCombo)

def ApplyHellIngress(Player, Enemy):
    if not (HellIngressEffect in Player.EffectList):
        Player.EffectList.append(HellIngressEffect)
        Player.EffectCDList.append(HellIngressCheck)
    Player.HellIngressCD = 20
    Player.HellIngressTimer = 20

def ApplyUnveiledGibbet(Player, Enemy):
    Player.AddGauge(-50) #Removing 50 Soul gauge
    Player.SoulReaverStack = 1

    #Since SoulReaver stack are removed if any other ability is done, we will add an effect that will check for that
    Player.EffectList.append(SoulReaverEffect)

def ApplySlice(Player, Enemy):
    if not (SliceCombo in Player.EffectList) : Player.EffectList.append(SliceCombo)
    Player.AddGauge(10) #add 10 Soul Gauge

def ApplyCommunio(Player, Enemy):
    #Ends Enshroud effect
    Player.AvatarTimer = 0
    Player.LemureGauge = 0
    Player.VoidShroudGauge = 0
    #Removing all Gauge related to Enshroud

def ApplyLemureSlice(Player, Enemy):
    Player.VoidShroudGauge -= 2 #Removing 2

def ApplyVoidReaping(Player, Enemy):
    if not (VoidReapingEffect in Player.EffectList) : 
        Player.EffectList.append(VoidReapingEffect)
        Player.EffectCDList.append(VoidReapingCheck)
    Player.LemureGauge -= 1
    Player.VoidShroudGauge = min(5, Player.VoidShroudGauge + 1) #Each Lemure used, we get a void shroud
    Player.VoidReapingTimer = 30

def ApplyCrossReaping(Player, Enemy):
    if not (CrossReapingEffect in Player.EffectList) : 
        Player.EffectList.append(CrossReapingEffect)
        Player.EffectCDList.append(CrossReapingCheck)
    Player.LemureGauge -= 1
    Player.VoidShroudGauge = min(5, Player.VoidShroudGauge + 1)
    Player.CrossReapingTimer = 30

def ApplyEnshroud(Player, Enemy):
    Player.AddShroud(-50) #Removing 50Shroud Gauge
    Player.EnshroudCD = 15
    Player.AvatarTimer = 30 #Avatar Timer
    Player.LemureGauge = 5 #Max Lemure


def ApplyPlentifulHarvest(Player, Enemy):
    Player.AddShroud(50)
    Player.ImmortalSacrificeStack = 0


def ApplyGibbet(Player, Enemy):
    if not Player.EnhancedGallows : 
        Player.EffectList.append(GibbetEffect) #Buffs next Gallows
        Player.EffectCDList.append(GibbetCheck)
        Player.EnhancedGallows = True
        #input("Giving enhanced Gallows")
    Player.AddShroud(10) #Add 10 ShroudGauge
    Player.GibbetEffectTimer = 60
    Player.SoulReaverStack -= 1 #Removing a stack

def ApplyGallows(Player, Enemy):
    if not Player.EnhancedGibbet : 
        Player.EffectList.append(GallowsEffect) #Buff next Gibbet
        Player.EffectCDList.append(GallowsCheck)
        Player.EnhancedGibbet = True
        #input("Giving enhanced Gibbet")
    Player.AddShroud(10) #Add 10 Shroud Gauge
    Player.GallowsEffectTimer = 60
    Player.SoulReaverStack -= 1 #Removing a stack

def ApplyBloodStalk(Player, Enemy):
    Player.AddGauge(-50) #Removing 50 Soul Gauge
    Player.SoulReaverStack = 1 #Reset to 1
    #Since SoulReaver stack are removed if any other ability is done, we will add an effect that will check for that
    Player.EffectList.append(SoulReaverEffect)


def ApplyGluttony(Player, Enemy):
    Player.AddGauge(-50) #Removing 50 Soul Gauge
    Player.GluttonyCD = 60
    Player.SoulReaverStack += 2
    #Since SoulReaver stack are removed if any other ability is done, we will add an effect that will check for that
    Player.EffectList.append(SoulReaverEffect)


def ApplyArcaneCircle(Player, Enemy):
    if not (ArcaneCircleBuff in Enemy.buffList):
        Enemy.buffList.append(ArcaneCircleBuff) #If not already in by another reaper
        Player.EffectCDList.append(ArcaneCircleCheck)
    #Will have to worry about if multiple reaper
    Player.ArcaneCircleCD = 120
    Player.ArcaneCircleTimer = 20
    Player.CircleOfSacrificeTimer = 5
    Player.BloodsownTimer = 6 #Cannot cast Plentiful harvest in that time

    #Here target is any player with the effect on, and Player is the Reaper casting it
    def CircleOfSacrificeEffet(Target, Spell):
        #This effect will be given to all players
        if Spell.GCD: #We will assume that the GCD will be either a Weaposkill or a spell
            Player.ImmortalSacrificeStack = min(8, Player.ImmortalSacrificeStack + 1)#Giving one stack
            Target.EffectToRemove.append(CircleOfSacrificeEffet)
            Target.EffectCDList.remove(CircleOfSacrificeCheck) #Removing check and effect, since max of 1 per player

    def CircleOfSacrificeCheck(Target, Enemy):
        if Player.CircleOfSacrificeTimer <= 0: #If no more time, we remove
            Target.EffectList.remove(CircleOfSacrificeEffet)
            Target.EffectToRemove.append(CircleOfSacrificeCheck)

    for player in Player.CurrentFight.PlayerList:
        #input("Given effect to : " + str(player))
        player.EffectList.append(CircleOfSacrificeEffet) #Giving the effect to all players in the fight
        player.EffectCDList.append(CircleOfSacrificeCheck)

                             # Only doing this if SavePreBakedAction is true
    #if Player.CurrentFight.SavePreBakedAction:
    #    fight = Player.CurrentFight
    #    history = buffPercentHistory(fight.TimeStamp, fight.TimeStamp + 20 , ArcaneCircleBuff.MultDPS)
    #    fight.PlayerList[fight.PlayerIDSavePreBakedAction].PercentBuffHistory.append(history)

def ApplySoulSlice(Player, Enemy):
    Player.AddGauge(50) #Adding 50 SoulGauge
    if Player.SoulSliceStack == 2: #If max stack, we add check
        Player.EffectCDList.append(SoulSliceStackCheck)
        Player.SoulSliceCD = 30
    Player.SoulSliceStack -= 1

def ApplyShadowOfDeath(Player, Enemy):
    if Player.DeathDesignTimer <= 0:
        #If not already applied
        Player.buffList.append(DeathDesignBuff) #10% bonus
        Player.EffectCDList.append(DeathDesignCheck)
    Player.DeathDesignTimer = min(60, Player.DeathDesignTimer + 30) #max of 60


def ApplyHarvestMoon(Player, Enemy):
    Player.Soulsow = False #Uses SoulSow

def ApplySoulsow(Player, Enemy):
    Player.Soulsow = True

#Effect

def SpinningScytheCombo(Player, Spell):
    if Spell.id == NightmareScythe.id:
        Spell.Potency += 60
        Player.AddGauge(10)
        Player.EffectToRemove.append(SpinningScytheCombo)

def HellIngressEffect(Player, Spell):
    if Spell.id == Harpe.id:
        Spell.CastTime = Lock #Insta Cast
        Player.HellIngressTimer = 0

def SoulReaverEffect(Player, Spell):
    #Will check if Spell is Gibbet or Gallows, if it is not we will remove SoulReaverStack
    if Spell.GCD and Spell.id != Gibbet.id and Spell.id != Gallows.id:
        #If not, we remove SoulReaverStack
        Player.SoulReaverStack = 0

def SliceCombo(Player, Spell):
    if Spell.id == WaxingSlice.id:
        if not (WaxingSliceCombo in Player.EffectList) : 
            Player.EffectList.append(WaxingSliceCombo)
        Player.EffectToRemove.append(SliceCombo)
        Spell.Potency += 240
        Player.AddGauge(10) #add 10 soul gauge

def WaxingSliceCombo(Player, Spell):
    if Spell.id == InfernalSlice.id:
        Spell.Potency += 320
        Player.EffectToRemove.append(WaxingSliceCombo)
        Player.AddGauge(10) #Add 10 soul gauge


def VoidReapingEffect(Player, Spell):
    if Spell.id == CrossReaping.id:
        Spell.Potency += 60
        Player.VoidReapingTimer = 0

def CrossReapingEffect(Player, Spell):
    if Spell.id == VoidReaping.id:
        Spell.Potency += 60
        Player.CrossReapingTimer = 0

def GibbetEffect(Player, Spell):
    if Spell.id == Gallows.id:
        #input("Buffing Gallows")
        Spell.Potency += 60
        Player.GibbetEffectTimer = 0
        Player.EnhancedGallows = False

def GallowsEffect(Player, Spell):
    if Spell.id == Gibbet.id:
        #input("Buffing Gibbet")
        Spell.Potency += 60
        Player.GallowsEffectTimer = 0
        Player.EnhancedGibbet = False

#check

def ArcaneCircleCheck(Player, Enemy):
    if Player.ArcaneCircleTimer <= 0:
        Enemy.buffList.remove(ArcaneCircleBuff)
        Player.EffectToRemove.append(ArcaneCircleCheck)

def HellIngressCheck(Player, Enemy):
    if Player.HellIngressTimer <= 0:
        Player.EffectList.remove(HellIngressEffect)
        Player.EffectToRemove.append(HellIngressCheck)

def VoidReapingCheck(Player, Enemy):
    if Player.VoidReapingTimer <= 0:
        Player.EffectList.remove(VoidReapingEffect)
        Player.EffectToRemove.append(VoidReapingCheck)

def CrossReapingCheck(Player, Enemy):
    if Player.CrossReapingTimer <= 0:
        Player.EffectList.remove(CrossReapingEffect)
        Player.EffectToRemove.append(CrossReapingCheck)

def GibbetCheck(Player, Enemy):
    if Player.GibbetEffectTimer <= 0:
        Player.EffectList.remove(GibbetEffect)
        Player.EffectToRemove.append(GibbetCheck)
        #input("Removing enhanced gallows")

def GallowsCheck(Player, Enemy):
    if Player.GallowsEffectTimer <= 0:
        Player.EffectList.remove(GallowsEffect)
        Player.EffectToRemove.append(GallowsCheck)
        #input("Removing enhanced gibbet")

def SoulSliceStackCheck(Player, Enemy):
    if Player.SoulSliceCD <= 0:
        if Player.SoulSliceStack == 1:
            Player.EffectToRemove.append(SoulSliceStackCheck)
        else:
            Player.SoulSliceCD = 30
        Player.SoulSliceStack += 1

def DeathDesignCheck(Player, Enemy):
    if Player.DeathDesignTimer <= 0:
        Player.buffList.remove(DeathDesignBuff)
        Player.EffectToRemove.append(DeathDesignCheck)

#GCD

#Combo Actions
Slice = ReaperSpell(24373, True, Lock, 2.5, 320, ApplySlice, [], True, type = 2)
WaxingSlice = ReaperSpell(24374, True, Lock, 2.5, 160, empty, [], True, type = 2)
InfernalSlice = ReaperSpell(24375, True, Lock, 2.5, 180, empty, [], True, type = 2)

#Other GCD
Soulsow = ReaperSpell(24387, True, 5, 2.5, 0, ApplySoulsow, [SoulsowRequirement], False, type = 1)
HarvestMoon = ReaperSpell(24388, True, Lock, 2.5, 600, ApplyHarvestMoon, [HarvestMoonRequirement], False, type = 1)
Harpe = ReaperSpell(24386, True, 1.3, 2.5, 300, empty, [], False, type = 1)
ShadowOfDeath = ReaperSpell(24378, True, Lock, 2.5, 300, ApplyShadowOfDeath, [], True, type = 2)
SoulSlice = ReaperSpell(24380, True, Lock, 2.5, 460, ApplySoulSlice, [SoulSliceRequirement], True, type = 2)
Gibbet = ReaperSpell(24382, True, Lock, 2.5, 460, ApplyGibbet, [GibbetRequirement], True, type = 2)
Gallows = ReaperSpell(24383, True, Lock, 2.5, 460, ApplyGallows, [GibbetRequirement], True, type = 2) #shares same requirement as Gibbet
PlentifulHarvest = ReaperSpell(24385, True, Lock, 2.5, 720, ApplyPlentifulHarvest, [PlentifulHarvestRequirement], True, type = 2)
VoidReaping = ReaperSpell(24395, True, Lock, 1.5, 460, ApplyVoidReaping, [VoidReapingRequirement], True, type = 2)
CrossReaping = ReaperSpell(24396, True, Lock, 1.5, 460, ApplyCrossReaping, [VoidReapingRequirement], True, type = 2) #Same Requriement as VoidReaping
Communio = ReaperSpell(24398, True, 1.3, 2.5, 1100, ApplyCommunio, [CommunioRequirement], False, type = 1)

#oGCD
ArcaneCircle = ReaperSpell(24405, False, Lock, 0, 0, ApplyArcaneCircle, [ArcaneCircleRequirement], False)
Gluttony = ReaperSpell(24393, False, Lock, 0, 520, ApplyGluttony, [GluttonyRequirement], False)
Enshroud = ReaperSpell(24394, False, Lock, 0, 0, ApplyEnshroud, [EnshroudRequirement], False)
LemureSlice = ReaperSpell(24399, False, Lock, 0, 240, ApplyLemureSlice, [LemureSliceRequirement], False)
BloodStalk = ReaperSpell(24389, False, Lock, 0, 340, ApplyBloodStalk, [BloodStalkRequirement], False)
UnveiledGibbet = ReaperSpell(24390, False, Lock, 0, 400, ApplyUnveiledGibbet, [UnveiledGibbetRequirement], False)
UnveiledGallows = ReaperSpell(24391, False, Lock, 0, 400, ApplyUnveiledGibbet, [UnveiledGallowsRequirement], False) #Shares effect with Gibbet
HellIngress = ReaperSpell(24402, False, Lock, 0, 0, ApplyHellIngress, [HellIngressRequirement], False)
ArcaneCrest = ReaperSpell(24404, False, 0, 0, 0, ApplyArcaneCrest, [ArcaneCrestRequirement], False)
#AOE GCD
SpinningScythe = ReaperSpell(24376, True, 0, 2.5, 140, ApplySpinningScythe, [], True, type = 2)
NightmareScythe = ReaperSpell(24377, True, 0, 2.5, 120, empty, [], True, type = 2)
WhorlOfDeath = ReaperSpell(24379, True, 0, 2.5, 100, ApplyShadowOfDeath, [], False, type = 2)
Guillotine = ReaperSpell(24384, False, 0, 0, 200, ApplyGuillotine, [GibbetRequirement], True, type = 2) #AOE version of Gallows/Gibbet
#AOE oGCD
GrimSwathe = ReaperSpell(24392, False, 0, 0, 140, ApplyBloodStalk, [BloodStalkRequirement], False) #AOE version of bloodstalk
SoulScythe = ReaperSpell(24381, False, 0, 0, 180, ApplySoulSlice, [SoulSliceRequirement], False) #AOE version of SoulSlice
GrimReaping = ReaperSpell(24397, True, 0, 2.5,200, ApplyGrimReaping, [VoidReapingRequirement], False)
LemureScythe = ReaperSpell(24400, False, 0, 0, 100, ApplyLemureSlice, [LemureSliceRequirement], False) #AOE version of Lemure Slice
#buff
DeathDesignBuff = buff(1.1,name="Death Design", isDebuff = True)
ArcaneCircleBuff = buff(1.03,name="AC")

ReaperAbility = {
24373 : Slice,
24374 : WaxingSlice,
24375 : InfernalSlice,
24378 : ShadowOfDeath,
24380 : SoulSlice,
24382 : Gibbet,
24383 : Gallows,
24376 : SpinningScythe,
24377 : NightmareScythe,
24379 : WhorlOfDeath,
24381 : SoulScythe,
24384 : Guillotine,
24389 : BloodStalk,
24390 : UnveiledGibbet,
24391 : UnveiledGallows,
24392 : GrimSwathe,
24393 : Gluttony,
24387 : Soulsow,
24388 : HarvestMoon,
24385 : PlentifulHarvest,
24394 : Enshroud,
24395 : VoidReaping,
24396 : CrossReaping,
24397 : GrimReaping,
24399 : LemureSlice,
24400 : LemureScythe,
24398 : Communio,
24404 : ArcaneCrest,
24405 : ArcaneCircle,
24402 : HellIngress,
24401 : HellIngress,
24386 : Harpe
}


"""
This file will contain model of return request from the API.
"""

from pydantic import BaseModel
from typing import List, Dict
class failedRequirementEventOut(BaseModel):
    timeStamp : float
    playerID : int
    requirementName : str
    additionalInfo : str
    fatal : bool

class fightInfoOut(BaseModel):
    fightDuration : float
    maxfightDuration : float
    fightname : str | None = "SimulatedFight"
    TeamCompositionBonus : float
    failedRequirementEventList : List[failedRequirementEventOut]
    Success : bool | str = "Unknown"

class GraphInfoOut(BaseModel):
    value : float | int
    name : float | int

class PlayerInfoOut(BaseModel):
    JobName : str
    ExpectedDPS : float
    PotencyPerSecond : float
    TotalDamage : float
    TotalPotency : float
    numberOfGCD : int
    ProcInfo : Dict
    GraphInfoDPS : List[GraphInfoOut] | List[None]
    GraphInfoPPS : List[GraphInfoOut] | List[None]

class dataOut(BaseModel):
    fightInfo : fightInfoOut
    PlayerList : List[PlayerInfoOut]

class SimulateFightOut(BaseModel):
    data : dataOut
    class Config:
        orm_mode = True





from ffxivcalc.Jobs.Base_Spell import buff, empty, DOTSpell, ManaRequirement
from ffxivcalc.Jobs.Tank.Tank_Spell import BigMitRequirement, ApplyBigMit, PaladinSpell
from ffxivcalc.helperCode.exceptions import InvalidTarget
from ffxivcalc.Jobs.Player import Shield, MitBuff
import copy
Lock = 0


#Requirement

def GoringBladeRequirement(Player, Spell):
    return Player.GoringBladeCD <= 0, Player.GoringBladeCD

def BulwarkRequirement(Player, Spell):
    return Player.BulwarkCD <= 0, Player.BulwarkCD

def FightOrFlightRequirement(Player, Spell):
    return Player.FightOrFlightCD <= 0, Player.FightOrFlightCD

def ExpiacionRequirement(Player, Spell):
    return Player.ExpiacionCD <= 0, Player.ExpiacionCD

def InterveneRequirement(Player, Spell):
    return Player.InterveneStack > 0, Player.InterveneCD

def AtonementRequirement(Player, Spell):
    return Player.SwordOathStack > 0, -1

def CircleScornRequirement(Player, Spell):
    return Player.CircleScornCD <= 0, Player.CircleScornCD

def BladeFaithRequirement(Player, Spell):
    return Player.BladeFaith, -1

def BladeTruthRequirement(Player, Spell):
    return Player.BladeTruth, -1

def BladeValorRequirement(Player, Spell):
    return Player.BladeValor, -1

def ConfettiRequirement(Player, Spell):
    return Player.RequestACat, -1

def RequestACatRequirement(Player, Spell):
    return Player.RequestACatCD <= 0, Player.RequestACatCD

def DivineVeilRequirement(Player, Spell):
    return Player.DivineVeilCD <= 0, Player.DivineVeilCD

def SheltronRequirement(Player, Spell):
    return Player.OathGauge >= 50, -1

def HolySheltronRequirement(Player, Spell):
    return Player.HolySheltronCD <= 0, Player.HolySheltronCD

def CoverRequirement(Player, Spell):
    return Player.CoverCD <= 0, Player.CoverCD

def InterventionRequirement(Player, Spell):
    return Player.InterventionCD <= 0, Player.InterventionCD

def HallowedGroundRequirement(Player, Spell):
    return Player.HallowedGroundCD <= 0, Player.HallowedGroundCD

#Apply

def ApplyHallowedGround(Player, Enemy):
    Player.HallowedGroundCD = 420

    Player.InvulnTimer = 10

def ApplyBulwark(Player, Enemy):
    Player.BulwarkCD = 90
    # TODO: clarify block vs normal mit
    # Block has no affect on dots.
    BulwarkMit = MitBuff(0.8, 10, Player)

    Player.MitBuffList.append(BulwarkMit)

def ApplyHolySheltron(Player, Enemy):
    Player.HolySheltronCD = 5
    Player.OathGauge -= 50

    
    # Gives 20% from block for 8 sec and
    # 15% mit for 4 seconds
    KnightResolveMit = MitBuff(0.85, 4, Player)
    HolySheltronMit = MitBuff(0.8, 8, Player)

    Player.MitBuffList.append(KnightResolveMit)
    Player.MitBuffList.append(HolySheltronMit)

def ApplyCover(Player, Enemy):
    Player.CoverCD = 120
    Player.OathGauge -= 50

def ApplyIntervention(Player, Enemy):
    Player.InterventionCD = 5
    Player.OathGauge -= 50

    

    

def ApplyDivineVeil(Player, Enemy):
    # For now will be assumed the buff is given as soon as its 
    # casted
    Player.DivineVeilCD = 90

    # Gives a shield to every player except the PLD
    # equal to 10% and cures every player by 400 potency

    shield_value = int(Player.MaxHP * 0.1)

    for player in Player.CurrentFight.PlayerList:
        # Giving every player that is not the PLD the divine veil shield
        if player != Player : player.ShieldList.append(Shield(shield_value, 30, player))
        

    



def ApplyTotalEclipse(Player, Enemy):
    if not (TotalEclipseCombo in Player.EffectList) : Player.EffectList.append(TotalEclipseCombo)

def ApplyFightOrFlight(Player, Enemy):
    Player.FightOrFlightTimer = 20
    Player.FightOrFlightCD = 60
    Player.buffList.append(FightOrFlightBuff)
    Player.EffectCDList.append(FightOrFlightCheck)

def ApplyExpiacion(Player, Enemy):
    Player.Mana += 10000 #Will have to double check that lol >.>
    Player.ExpacionCD = 30

def ApplyIntervene(Player, Enemy):
    if Player.InterveneStack == 2:
        Player.EffectCDList.append(InterveneStackCheck)
        Player.InterveneCD = 30
    Player.InterveneStack -= 1

def ApplyAtonement(Player, Enemy):
    Player.SwordOathStack -= 1

def ApplyCircleScorn(Player, Enemy):
    #Since DOT has 30 sec CD and 15 sec duration, we don't have to check if already applied
    Player.CircleScornCD = 30
    Player.CircleScornDOT = copy.deepcopy(CircleScornDOT)
    Player.DOTList.append(Player.CircleScornDOT)
    Player.CircleScornTimer = 15
    Player.EffectCDList.append(CircleScornDOTCheck)

def ApplyRequestACat(Player, Enemy):
    Player.RequestACatCD = 60
    Player.RequestACat = True
    Player.RequestACatStack = 4
    Player.EffectList.append(RequestACatEffect)
    Player.EffectCDList.append(RequestACatCheck)

def ApplyBladeFaith(Player, Enemy):
    Player.BladeFaith = False
    Player.BladeTruth = True

def ApplyBladeTruth(Player, Enemy):
    Player.BladeTruth = False
    Player.BladeValor = True

def ApplyBladeValor(Player, Enemy):
    Player.BladeValor = False


def ApplyConfetti(Player, Enemy):
    Player.BladeFaith = True


def ApplyFastBlade(Player, Enemy):
    if not (FastBladeCombo in Player.EffectList) : 
        Player.EffectList.append(FastBladeCombo)

def ApplyGoringBlade(Player, Enemy):
    Player.GoringBladeCD = 60 * Player.WeaponskillReduction # Cuz is weaponskill

def DivineMightEffect(Player, Spell):
    """Allows next holy spirit or holy circle to be cast immediately with increased potency.
    """
    if Spell.id == HolySpirit.id:
        Spell.Potency += 100
        Spell.CastTime = 0
        Player.EffectToRemove.append(DivineMightEffect)
    elif Spell.id == HolyCircle.id:
        Spell.Potency += 100
        Spell.CastTime = 0
        Player.EffectToRemove.append(DivineMightEffect)

def RequestACatEffect(Player, Spell):

    match Spell.id:
        case HolySpirit.id:
                             # Divine might has prio over this.
            if not DivineMightEffect in Player.EffectList:
                Spell.Potency += 300
                Spell.CastTime = 0
                Player.RequestACatStack -= 1
        case HolyCircle.id:
                             # Divine might has prio over this.
            if not DivineMightEffect in Player.EffectList:
                Spell.Potency += 170
                Spell.CastTime = 0
                Player.RequestACatStack -= 1
        case Clemency.id:
            Spell.CastTime = 0
            Player.RequestACatStack -= 1
    #Confiteor combo
        case Confetti.id | BladeFaith.id | BladeTruth.id | BladeValor.id:
            Spell.Potency += 500
            Player.RequestACatStack -= 1

    



#Combo Action

def TotalEclipseCombo(Player, Spell):
    if Spell.id == Prominence.id:
        Spell.Potency += 70
        Player.EffectToRemove.append(TotalEclipseCombo)
        # also gives Divine might
        Player.EffectList.append(DivineMightEffect)


def FastBladeCombo(Player, Spell):
    if Spell.id == RiotBlade.id:
        Spell.Potency += 160
        Player.EffectToRemove.append(FastBladeCombo)
        if not (RiotBladeCombo in Player.EffectList) : Player.EffectList.append(RiotBladeCombo)

def RiotBladeCombo(Player, Spell):
    if Spell.id == RoyalAuthority.id:
        Spell.Potency += 260
        Player.SwordOathStack += 3
        Player.EffectToRemove.append(RiotBladeCombo)
        # also gives Divine might
        Player.EffectList.append(DivineMightEffect)

    """
    Old goring blade
    elif Spell.id == GoringBlade.id:
        #Apply dot

        Player.GoringDOT = copy.deepcopy(GoringDOT)
        Player.EffectCDList.append(GoringDOTCheck)
        Player.DOTList.append(Player.GoringDOT)
        Player.ValorDOTTimer = 0 #Remove other dot if it exists
        Player.GoringDOTTimer = 21
        Spell.Potency += 150
        Player.EffectToRemove.append(RiotBladeCombo)
    """


#Check

def DivineVeilCheck(Player, Enemy):
    if Player.DivineVeilTimer <= 0:
        Player.EffectToRemove.append(DivineVeilCheck)

def RequestACatCheck(Player, Enemy):
    if Player.RequestACatStack == 0:
        Player.RequestACat = False
        Player.EffectList.remove(RequestACatEffect)
        Player.EffectToRemove.append(RequestACatCheck)
        Player.BladeFaith = True


def FightOrFlightCheck(Player, Enemy):
    if Player.FightOrFlightTimer <= 0:
        Player.buffList.remove(FightOrFlightBuff)
        Player.EffectToRemove.append(FightOrFlightCheck)

def InterveneStackCheck(Player, Enemy):
    if Player.InterveneCD <= 0:
        if Player.InterveneStack == 1:
            Player.EffectToRemove.append(InterveneStackCheck)
        else:
            Player.InterveneCD = 30
        Player.InterveneStack += 1

def CircleScornDOTCheck(Player, Enemy):
    if Player.CircleScornTimer <= 0:
        Player.DOTList.remove(Player.CircleScornDOT)
        Player.CircleScornDOT = None
        Player.EffectToRemove.append(CircleScornDOTCheck)
"""
Post 6.3
def GoringDOTCheck(Player, Enemy):
    if Player.GoringDOTTimer <= 0:
        Player.DOTList.remove(Player.GoringDOT)
        Player.EffectToRemove.append(GoringDOTCheck)
        Player.GoringDOT = None

def ValorDOTCheck(Player, Enemy):
    if Player.ValorDOTTimer <= 0:
        Player.DOTList.remove(Player.ValorDOT)
        Player.EffectToRemove.append(ValorDOTCheck)
        Player.ValorDOT = None
"""

#Combo action

FastBlade = PaladinSpell(9, True, Lock, 2.5, 200, 0, ApplyFastBlade, [], True, type = 2)
RiotBlade = PaladinSpell(15, True, Lock, 2.5, 140, 0, empty, [], True, type = 2)
RoyalAuthority = PaladinSpell(3539, True, Lock, 2.5, 140,0, empty, [], True, type = 2)
GoringBlade = PaladinSpell(3538, True, Lock, 2.5, 700, 0, ApplyGoringBlade, [GoringBladeRequirement], True, type = 2)
"""
Post 6.3 goring blade
GoringBlade = PaladinSpell(3538, True, Lock, 2.5, 100, 0, empty, [], True, type = 2)
GoringDOT = DOTSpell(-5, 65, True)
"""
#Confiteor Combo Action

Confetti = PaladinSpell(16459, True, Lock, 2.5, 420, 1000, ApplyConfetti, [ManaRequirement, ConfettiRequirement], False, type = 1) # >.>
BladeFaith = PaladinSpell(25748, True, Lock, 2.5, 220, 0, ApplyBladeFaith, [BladeFaithRequirement], False, type = 1)
BladeTruth = PaladinSpell(25749, True, Lock, 2.5, 320, 0, ApplyBladeTruth, [BladeTruthRequirement], False, type = 1)
BladeValor = PaladinSpell(25750, True, Lock, 2.5, 420, 0, ApplyBladeValor, [BladeValorRequirement], False, type = 1)
"""
Post 6.3
BladeValorDOT = DOTSpell(-11, 80, True)
"""
#GCD
HolySpirit = PaladinSpell(7384, True, 1.5, 2.5, 350, 1000, empty, [ManaRequirement], False, type = 1)
Atonement = PaladinSpell(16460, True, Lock, 2.5, 400, 0, ApplyAtonement, [AtonementRequirement], True, type = 1)
Clemency = PaladinSpell(3541, True, 1.5, 2.5, 0, 1000, empty, [ManaRequirement], False, type = 2)
ShieldLob = PaladinSpell(24, True, 0, 2.5, 100, 0, empty, [], True , type = 2)
#AOE GCD
HolyCircle = PaladinSpell(16458, True, 1.5, 2.5,100, 1000, empty, [ManaRequirement], False, type = 1)
TotalEclipse = PaladinSpell(7381, True, 0, 2.5, 100, 0, ApplyTotalEclipse, [], True, type = 2)
Prominence = PaladinSpell(16457, True, 0, 2.5, 100, 0, empty, [], True, type = 2)

#oGCD
RequestACat = PaladinSpell(7383, False, 0, Lock, 320, 0, ApplyRequestACat, [RequestACatRequirement], True) #I NEED ONE RIGHT NOW :x
CircleScorn = PaladinSpell(23, False, 0, Lock, 140, 0, ApplyCircleScorn, [CircleScornRequirement], True)
CircleScornDOT = DOTSpell(-6, 30, True)
Intervene = PaladinSpell(16461, False, 0, Lock, 150, 0, ApplyIntervene, [InterveneRequirement], True)
Expiacion = PaladinSpell(25747, False, 0, Lock, 450, 0, ApplyExpiacion, [ExpiacionRequirement], True)
FightOrFlight = PaladinSpell(20, False, 0, Lock, 0, 0, ApplyFightOrFlight, [FightOrFlightRequirement], True)

#Mitigation Actions
DivineVeil = PaladinSpell(3540, False, 0, 0, 0, 0, ApplyDivineVeil, [DivineVeilRequirement], False)
HolySheltron = PaladinSpell(25746, False, 0, 0, 0, 0, ApplyHolySheltron, [HolySheltronRequirement,SheltronRequirement], False)
Cover = PaladinSpell(27, False, 0, 0, 0, 0, ApplyCover, [SheltronRequirement, CoverRequirement], False)
HallowedGround = PaladinSpell(30, False, 0, 0, 0, 0, ApplyHallowedGround, [HallowedGroundRequirement], False)
Bulwark = PaladinSpell(11111, False, 0, 0, 0, 0, ApplyBulwark, [BulwarkRequirement], False)
Sentinel = PaladinSpell(17,False, 0, 0, 0, 0, ApplyBigMit, [BigMitRequirement], False)

def Intervention(Target):
    """This function returns a PLDSpell object corresponding to
    intervention.

    Args:
        Target (Player) : Target of the action
    
    """

    def Apply(Player, Spell):
        ApplyIntervention(Player, Spell)

        # Check if the target is valid

        if Target == Player: # Intervention cannot be used on the player itself
            raise InvalidTarget("Intervention", Player, Target, False, Target.playerID)

        # 10% for 8 sec. 10% for 4 secs and 10% (flat bonus)
        # if rampart or sentinel is up

        mit = 0.9

        if Player.BigMitTimer > 0 or Player.RampartTimer > 0:
            mit = 0.8
            # Rampart and/or Sentinel are used
        
        InterventionBuff = MitBuff(mit, 8, Target)
        KnightBuff = MitBuff(0.9, 4, Target)

        Target.MitBuffList.append(InterventionBuff)
        Target.MitBuffList.append(KnightBuff)


    Intervention = PaladinSpell(7382, False, 0, 0, 0, 0, Apply, [SheltronRequirement, InterventionRequirement], False)
    Intervention.TargetID = Target.playerID
    return Intervention




def PassageOfArms(time):
    #Function since we will be using it for a set time


    def PassageOfArmsRequirement(Player, Spell):
        return Player.PassageOfArmsCD <= 0, Player.PassageOfArmsCD

    def ApplyPassageOfArms(Player, Enemy):
        Player.PassageOfArmsCD = 120

        # Will assume every player gets 15% mit for 6 sec

        for player in Player.CurrentFight.PlayerList:
            player.MitBuffList.append(MitBuff(0.85, 5, player))


    return PaladinSpell(7385, False, time, time, 0, 0, ApplyPassageOfArms, [PassageOfArmsRequirement], False)
#buff
FightOrFlightBuff = buff(1.25,name="Fight or Flight")

PaladinAbility = {
30 : HallowedGround,
24 : ShieldLob,
16461 : Intervene,
20 : FightOrFlight,
7383 : RequestACat,
9 : FastBlade,
15 : RiotBlade,
3538 : GoringBlade,
3539 : RoyalAuthority,
16460 : Atonement,
7384 : HolySpirit,
16459 : Confetti,
25748 : BladeFaith,
25749 : BladeTruth,
25750 : BladeValor,
25747 : Expiacion,
7381 : TotalEclipse,
16457 : Prominence,
23 : CircleScorn,
16458 : HolyCircle,
17 : Sentinel,
25746 : HolySheltron,
7382 : Intervention,
27 : Cover,
3541 : Clemency,
3540 : DivineVeil,
7385 : PassageOfArms(1)
}


from ffxivcalc.Jobs.Base_Spell import Potion, Spell, empty
Lock = 0
class MeleeSpell(Spell):

    def __init__(self, id, GCD, CastTime, RecastTime, Potency, ManaCost, Effect, Requirement, type = 0):
        super().__init__(id, GCD, CastTime, RecastTime, Potency, ManaCost, Effect, Requirement, type = type)

#########################################
########## NINJA SPELL  #################
#########################################

class NinjaSpell(MeleeSpell):

    def __init__(self, id, GCD, CastTime, RecastTime, Potency, Effect, Requirement,Weaponskill, Ninjutsu, type = 0):
        super().__init__(id, GCD, CastTime, RecastTime, Potency, 0, Effect, Requirement, type = type)
        self.Weaponskill = Weaponskill
        self.Ninjutsu = Ninjutsu


#########################################
########## SAMURAI PLAYER ###############
#########################################

class SamuraiSpell(MeleeSpell):
    def __init__(self, id, GCD, CastTime, RecastTime, Potency, Effect, Requirement, KenkiCost, type = 0):
        super().__init__(id, GCD, CastTime, RecastTime, Potency, 0, Effect, Requirement, type = type)

        self.KenkiCost = KenkiCost
        self.Requirement += [KenkiRequirement]

def KenkiRequirement(Player, Spell): #By default present in Samurai spell requirements
    #input("Current Gauge is " + str(Player.KenkiGauge - Spell.KenkiCost))
    return Spell.KenkiCost <= Player.KenkiGauge, -1

#########################################
########## DRAGOON PLAYER ###############
#########################################

class DragoonSpell(MeleeSpell):
    def __init__(self, id, GCD, RecastTime, Potency, Effect, Requirement, Weaponskill, type = 0):
        super().__init__(id, GCD, Lock, RecastTime, Potency, 0, Effect, Requirement, type = type)

        self.Weaponskill = Weaponskill

#########################################
########## MONK SPELL ###############
#########################################

class MonkSpell(MeleeSpell):
    def __init__(self, id, GCD, RecastTime, Potency, Effect, Requirement, Weaponskill, MasterfulBlitz, type = 0):
        super().__init__(id, GCD, Lock, RecastTime, Potency, 0, Effect, Requirement, type = type)

        self.Weaponskill = Weaponskill
        self.MasterfulBlitz = MasterfulBlitz

#########################################
########## MONK SPELL ###############
#########################################

class ReaperSpell(MeleeSpell):
    def __init__(self, id, GCD, CastTime,RecastTime, Potency, Effect, Requirement, Weaponskill, type = 0):
        super().__init__(id, GCD, CastTime, RecastTime, Potency, 0, Effect, Requirement, type = type)

        self.Weaponskill = Weaponskill



#Class action spell

#Requirement

def SecondWindRequirement(Player, Spell):
    return Player.SecondWindCD <= 0, Player.SecondWindCD

def LegSweepRequirement(Player, Spell):
    return Player.LegSweepCD <= 0, Player.LegSweepCD

def BloodbathRequirement(Player, Spell):
    return Player.BloodbathCD <= 0, Player.BloodbathCD

def FeintRequirement(Player, Spell):
    return Player.FeintCD <= 0, Player.FeintCD

def ArmLengthRequirement(Player, Spell):
    return Player.ArmLengthCD <= 0, Player.ArmLengthCD

def TrueNorthRequirement(Player, Spell):
    return Player.TrueNorthStack > 0, Player.TrueNorthCD

#Apply
def ApplySecondWind(Player, Enemy):
    Player.SecondWindCD = 120

def ApplyLegSweep(Player, Enemy):
    Player.LegSweepCD = 40

def ApplyBloodbath(Player, Enemy):
    Player.BloodbathCD = 90

def ApplyFeint(Player, Enemy):
    Player.FeintCD = 90
    Enemy.Feint = True
    Enemy.FeintTimer = 10

def ApplyArmLength(Player, Enemy):
    Player.ArmLengthCD = 120

def ApplyTrueNorth(Player, Enemy):
    if Player.TrueNorthStack == 2:
        Player.EffectCDList.append(TrueNorthStackCheck)
        Player.TrueNorthCD = 45
    Player.TrueNorthStack -= 1


#Check

def TrueNorthStackCheck(Player, Enemy):
    if Player.TrueNorthCD <= 0:
        if Player.TrueNorthStack == 1:
            Player.EffectToRemove.append(TrueNorthStackCheck)
        else:
            Player.TrueNorthCD = 45
        Player.TrueNorthStack += 1
#Class Action (no effect as of now)
SecondWind = MeleeSpell(7541, False, Lock, 0, 0, 0, ApplySecondWind, [SecondWindRequirement])
LegSweep = MeleeSpell(7863, False, Lock, 0, 0, 0, ApplyLegSweep, [LegSweepRequirement])
Bloodbath = MeleeSpell(7542, False, Lock, 0, 0, 0, ApplyBloodbath, [BloodbathRequirement])
Feint = MeleeSpell(7549, False, Lock, 0, 0, 0, ApplyFeint, [FeintRequirement])
ArmLength = MeleeSpell(7548, False, Lock, 0, 0, 0, ApplyArmLength, [ArmLengthRequirement])
TrueNorth = MeleeSpell(7546, False, Lock, 0, 0, 0, ApplyTrueNorth, [TrueNorthRequirement])

# Limit Break actions

LB1Timer = 5.86
LB2Timer = 6.86
LB3Timer = 8.20

MeleeLB1 = MeleeSpell(1111, False,LB1Timer, LB1Timer, 0, 0, empty, [], type=3)
MeleeLB2 = MeleeSpell(1112, False,LB2Timer, LB2Timer, 0, 0, empty, [], type=3)
MeleeLB3 = MeleeSpell(1113, False,LB3Timer, LB3Timer, 0, 0, empty, [], type=3)


MeleeAbility = {
7541 : SecondWind,
7863 : LegSweep,
7542 : Bloodbath,
7549 : Feint,
7546 : TrueNorth,
7548 : ArmLength,
34590541 : Potion,  #STR
34590542 : Potion,   #DEX,
34592395 : Potion, # STR 7
34592396 : Potion, # DEX 7
34594159 : Potion, # STR 8
34594160 : Potion, # DEX pot ?
-2 : Potion,
1111 : MeleeLB1,
1112 : MeleeLB2,
1113 : MeleeLB3
}


#########################################
########## SCHOLAR PLAYER ###############
#########################################

from ffxivcalc.Jobs.Healer.Healer_Spell import ScholarSpell
from ffxivcalc.Jobs.Player import HealingBuff, MitBuff

from ffxivcalc.Jobs.Base_Spell import DOTSpell, empty, ManaRequirement, buffHistory
import copy
Lock = 0.75

def AetherflowRequirement(Player, Spell):
    return Player.AetherFlowCD <= 0, Player.AetherFlowCD

def ChainStratagemRequirement(Player, Spell):
    return Player.ChainStratagemCD <= 0, Player.ChainStratagemCD

def AetherStackRequirement(Player, Spell):
    return Player.AetherFlowStack > 0, -1

def DissipationRequirement(Player, Spell):
    return Player.DissipationCD <= 0 and Player.SummonTimer <= 0, Player.DissipationCD #Cannot use while summoned

def ExpedientRequirement(Player, Spell):
    return Player.ExpedientCD <= 0, Player.ExpedientCD

def ProtractionRequirement(Player, Spell):
    return Player.ProtractionCD <= 0, Player.ProtractionCD

def RecitationRequirement(Player, Spell):
    return Player.RecitationCD <= 0, Player.RecitationCD

def EmergencyTacticRequirement(Player, Spell):
    return Player.EmergencyTacticCD <= 0, Player.EmergencyTacticCD

def DeploymentTacticRequirement(Player, Spell):
    return Player.DeploymentTacticCD <= 0, Player.DeploymentTacticCD

def IndomitabilityRequirement(Player, Spell):
    return Player.IndomitabilityCD <= 0, Player.IndomitabilityCD

def LustrateRequirement(Player, Spell):
    return Player.LustrateCD <= 0, Player.LustrateCD

def SacredSoilRequirement(Player, Spell):
    return Player.SacredSoilCD <= 0, Player.SacredSoilCD

def ExcogitationRequirement(Player, Spell):
    return Player.ExcogitationCD <= 0, Player.ExcogitationCD

def AdloquiumRequirement(Player, Spell):
    if Player.Recitation:
        return True, -1 #if we have it
    else:
        return ManaRequirement(Player, Spell) #Else return ManaRequirement

def AetherHealRequirement(Player, Spell):
    return Player.AetherFlowStack > 0 or Player.Recitation, -1

def SummonSeraphRequirement(Player, Spell):
    return Player.SummonSeraphCD <= 0, Player.SummonSeraphCD

def ConsolationRequirement(Player, Spell):
    return Player.ConsolationStack > 0, -1

def FeyBlessingRequirement(Player, Spell):
    return Player.FeyBlessingCD <= 0, Player.FeyBlessingCD

def FeyIlluminationRequirement(Player, Spell):
    return Player.FeyIlluminationCD <= 0, Player.FeyIlluminationCD

def WhisperingDawnRequirement(Player, Spell):
    return Player.WhisperingDawnCD <= 0, Player.WhisperingDawnCD

#====================================

def ApplyFeyBlessing(Player, Enemy):
    Player.FeyBlessingCD = 60

def ApplyFeyIllumination(Player, Enemy):
    Player.FeyIlluminationCD = 120

    for player in Player.CurrentFight.PlayerList:
        player.AddHealingBuff(HealingBuff(1.1, 20, player, GivenHealBuff=False, BuffName="FeyIllumination"))
        player.AddMitBuff(MitBuff(0.95, 20, player, MagicMit=True, BuffName="FeyIllumination"))

def ApplyWhisperingDawn(Player, Enemy):
    Player.WhisperingDawnCD = 60

def Apply(Player, Enemy):
    Player.CD = 0
    Player.AetherFlowStack -= 1

def ApplyConsolation(Player, Enemy):
    Player.ConsolationStack -= 1

def ApplySummonSeraph(Player, Enemy):
    Player.SummonTimer = 22
    Player.SummonSeraphCD = 120
    Player.ConsolationStack = 2

def ApplyIndomitability(Player, Enemy):
    Player.IndomitabilityCD = 30
    if Player.Recitation : Player.Recitation = False
    else: Player.AetherFlowStack -= 1

def ApplyLustrate(Player, Enemy):
    Player.LustrateCD = 1
    Player.AetherFlowStack -= 1

def ApplySacredSoil(Player, Enemy):
    Player.SacredSoilCD = 30
    Player.AetherFlowStack -= 1

    for player in Player.CurrentFight.PlayerList:
        player.AddMitBuff(MitBuff(0.9, 15, player, BuffName="SacredSoil"))

def ApplyExcogitation(Player, Enemy):
    Player.ExcogitationCD = 45
    if Player.Recitation : Player.Recitation = False
    else: Player.AetherFlowStack -= 1

def ApplyAdloquium(Player, Enemy):
    if Player.Recitation : Player.Recitation = False

def ApplyExpedient(Player, Enemy):
    Player.ExpedientCD = 120

    for player in Player.CurrentFight.PlayerList:
        player.AddMitBuff(MitBuff(0.9, 20, player, BuffName="Expedient"))

def ApplyProtraction(Player, Enemy):
    Player.ProtractionCD = 60

def ApplyRecitation(Player, Enemy):
    Player.RecitationCD = 90
    Player.Recitation = True

def ApplyEmergencyTactic(Player, Enemy):
    Player.EmergencyTacticCD = 15

def ApplyDeploymentTactic(Player, Enemy):
    Player.DeploymentTacticCD = 90
        
def ApplyBiolysis(Player, Enemy):
    if (not (Player.Biolysis in Player.DOTList)):
        Player.Biolysis = copy.deepcopy(BiolysisDOT)
        Player.DOTList.append(Player.Biolysis)
        Player.EffectCDList.append(CheckBiolysis)
    else : Player.Biolysis.resetBuffSnapshot() # If already applied reset snapshot
    Player.BiolysisTimer = 30

def ApplyAetherflow(Player, Enemy):
    Player.Mana = min(10000, Player.Mana + 2000)
    Player.AetherFlowStack = 3
    Player.AetherFlowCD = 60

def ApplyChainStratagem(Player, Enemy):
    Player.ChainStratagemCD = 120
    Player.ChainStratagemTimer = 15
    Enemy.ChainStratagem = True
    Player.EffectCDList.append(CheckChainStratagem)

                                     # Only doing this if SavePreBakedAction is true
    #if Player.CurrentFight.SavePreBakedAction:
    #    fight = Player.CurrentFight
    #    history = buffHistory(fight.TimeStamp, fight.TimeStamp + 15)
    #    fight.PlayerList[fight.PlayerIDSavePreBakedAction].ChainStratagemHistory.append(history)

def ApplyEnergyDrain(Player, Enemy):
    Player.EnergyDrainCD = 1
    Player.AetherFlowStack -= 1

def ApplyDissipation(Player, Enemy):
    Player.DissipationCD = 180
    Player.AetherFlowStack = 3
    Player.GivenHealBuffList.append(HealingBuff(1.2, 30, Player, BuffName="Dissipation"))

#===================================

def CheckChainStratagem(Player, Enemy):
    if Player.ChainStratagemTimer <= 0:
        Player.ChainStratagemTimer = 0
        Enemy.ChainStratagem = False
        Player.EffectToRemove.append(CheckChainStratagem)

def CheckBiolysis(Player, Enemy):
    if Player.BiolysisTimer <= 0 : 
        Player.DOTList.remove(Player.Biolysis)
        Player.Biolysis = None
        Player.EffectToRemove.append(CheckBiolysis)




#DamageGCD
Broil = ScholarSpell(25865, True, 1.5, 2.5, 295,  400, empty, [ManaRequirement], type = 1)
Ruin = ScholarSpell(17870, True, 0, 2.5, 220,  300, empty, [ManaRequirement], type = 1)
Biolysis = ScholarSpell(16540, True, 0, 2.5, 0, 300, ApplyBiolysis, [ManaRequirement], type = 1)
BiolysisDOT = DOTSpell(-4, 70, False)
ArtOfWar = ScholarSpell(25866, True, 0, 2.5, 165, 400, empty, [ManaRequirement], type = 1)#AOE

#HealGCD
Succor = ScholarSpell(186, True, 2, 2.5, 200, 1000, ApplyAdloquium, [AdloquiumRequirement], type = 1, AOEHeal=True) ##This action has apply and requirement because of recitation. 160% shield
Adloquium = ScholarSpell(185, True, 2, 2.5, 300, 1000, ApplyAdloquium, [AdloquiumRequirement], type = 1, TargetHeal=True) #This action has apply and requirement because of recitation. 180% shield
Physick = ScholarSpell(190, True, 1.5, 2.5, 450, 400, empty, [ManaRequirement], type = 1, TargetHeal=True)
SummonEos = ScholarSpell(29, True, 1.5, 2.5, 0, 200, empty, [ManaRequirement], type = 1)
Resurrection = ScholarSpell(173, True, 8, 2.5, 0, 2400, empty, [ManaRequirement])
#Damage oGCD
ChainStratagem = ScholarSpell(7436, False, 0, Lock, 0, 0, ApplyChainStratagem, [ChainStratagemRequirement])
EnergyDrain = ScholarSpell(167, False, 0, Lock, 100, 0, ApplyEnergyDrain, [AetherStackRequirement])
Aetherflow = ScholarSpell(166, False, 0, Lock, 0, 0, ApplyAetherflow, [AetherflowRequirement])
Dissipation = ScholarSpell(3587, False, 0, Lock, 0, 0, ApplyDissipation, [DissipationRequirement])
#Heal oGCD
Expedient = ScholarSpell(25868, False, 0, 0, 0, 0, ApplyExpedient, [ExpedientRequirement])
Protraction = ScholarSpell(25867, False, 0, 0, 0, 0, ApplyProtraction, [ProtractionRequirement])
Recitation = ScholarSpell(16542, False, 0, 0, 0, 0, ApplyRecitation, [RecitationRequirement])
EmergencyTactic = ScholarSpell(3586, False, 0, 0, 0, 0, ApplyEmergencyTactic, [EmergencyTacticRequirement])
DeploymentTactic = ScholarSpell(3585, False, 0, 0, 0, 0, ApplyDeploymentTactic, [DeploymentTacticRequirement])

#AetherFlow Heal spell
Excogitation = ScholarSpell(7434, False, 0, 0, 0, 0, ApplyExcogitation, [ExcogitationRequirement,AetherHealRequirement]) #Can be used by recitation
SacredSoil = ScholarSpell(188, False, 0, 0, 0, 0, ApplySacredSoil, [SacredSoilRequirement,AetherflowRequirement])
Lustrate = ScholarSpell(189, False, 0, 0, 600, 0, ApplyLustrate, [LustrateRequirement,AetherflowRequirement], TargetHeal=True)
Indomitability = ScholarSpell(3583, False, 0, 0, 400, 0, ApplyIndomitability, [IndomitabilityRequirement,AetherHealRequirement], AOEHeal=True)#Can be used by recitation

#Fey Healing
Consolation = ScholarSpell(16546, False, 0, 0, 0, 0, ApplyConsolation, [ConsolationRequirement])
SummonSeraph = ScholarSpell(16545, False, 0, 0, 250, 0, ApplySummonSeraph, [SummonSeraphRequirement], AOEHeal=True) # 100% shield
FeyBlessing = ScholarSpell(16543, False, 0, 0, 320, 0, ApplyFeyBlessing, [FeyBlessingRequirement], AOEHeal=True)
Aetherpact = ScholarSpell(7437, False, 0, 0, 0, 0, empty, []) #No requirement, since 3 sec cd, so not really worth it imo
DissolveUnion = ScholarSpell(7869, False, 0, 0, 0, 0, empty, [])
FeyIllumination = ScholarSpell(16538, False, 0, 0, 0, 0, ApplyFeyIllumination, [FeyIlluminationRequirement])
WhisperingDawn = ScholarSpell(16537, False, 0, 0, 0, 0, ApplyWhisperingDawn, [WhisperingDawnRequirement])

ScholarAbility = {
166 : Aetherflow, 
167 : EnergyDrain, 
185 : Adloquium,  
186 : Succor, 
188 : SacredSoil, 
189 : Lustrate, 
190 : Physick, 
3583 : Indomitability, 
3585 : DeploymentTactic, 
3586 : EmergencyTactic,
3587 : Dissipation, 
7434 : Excogitation, 
7436 : ChainStratagem, 
7437 : Aetherpact, 
16537 : WhisperingDawn, 
16538 : FeyIllumination, 
16542 : Recitation, 
16543 : FeyBlessing, 
16545 : SummonSeraph,
17215 : SummonEos, 
25868 : Expedient, 
25867 : Protraction, 
25866 : ArtOfWar, 
25865 : Broil, 
17870 : Ruin, 
16540 : Biolysis,
16546 : Consolation,
173 : Resurrection,
7869 : DissolveUnion
}


from ffxivcalc.Jobs.Base_Spell import DOTSpell, ManaRequirement, WaitAbility, buff, empty, buffPercentHistory
from ffxivcalc.Jobs.Caster.Caster_Spell import SummonerSpell
import copy

from ffxivcalc.Jobs.Player import Pet

Lock = 0

#Requirement

def RekindleRequirement(Player, Spell):
    return True, -1

def Ruin4Requirement(Player, Spell):
    return Player.FurtherRuin, -1

def TitanRequirement(Player, Spell):
    return Player.TitanGem, -1

def IfritRequirement(Player, Spell):
    return Player.IfritGem, -1

def GarudaRequirement(Player, Spell):
    return Player.GarudaGem, -1

def TopazRequirement(Player, Spell):
    return Player.TitanStack > 0, -1

def MountainRequirement(Player, Spell):
    return Player.TitanSpecial, -1

def RubyRequirement(Player, Spell):
    return Player.IfritStack > 0, -1

def CycloneRequirement(Player, Spell):
    return Player.IfritSpecial, -1

def StrikeRequirement(Player, Spell):
    return Player.IfritSpecialCombo, -1

def EmeraldRequirement(Player, Spell):
    return Player.GarudaStack > 0, -1

def SlipstreamRequirement(Player, Spell):
    return Player.GarudaSpecial, -1

def SummonRequirement(Player, Spell):
    return Player.SummonCD <= 0, Player.SummonCD

def FoFRequirement(Player, Spell):
    return Player.FirebirdTrance, -1

def AstralImpulseRequirement(Player, Spell):
    return Player.BahamutTrance, -1

def EnkindleRequirement(Player, Spell):
    return Player.Enkindle, -1

def DeathflareRequirement(Player, Spell):
    return Player.Deathflare, -1

def EnergyDrainRequirement(Player, Spell):
    return Player.EnergyDrainCD <= 0, Player.EnergyDrainCD

def FesterRequirement(Player, Spell):
    return Player.AetherflowGauge > 0, -1

def SearingLightRequirement(Player, Spell):
    return Player.SearingLightCD <= 0, Player.SearingLightCD

#Apply

def ApplyRuin4(Player, Enemy):
    Player.FurtherRuin = False

def ApplyTitan(Player, Enemy):
    Player.resetPrimalEffect()
    Player.TitanGem = False
    Player.TitanStack = 4

def ApplyIfrit(Player, Enemy):
    Player.resetPrimalEffect()
    Player.IfritGem = False
    Player.IfritStack = 2
    Player.IfritSpecial = True

def ApplyGaruda(Player, Enemy):
    Player.resetPrimalEffect()
    Player.GarudaGem = False
    Player.GarudaStack = 4
    Player.GarudaSpecial = True

def ApplyTopaz(Player, Enemy):
    Player.TitanStack -= 1
    Player.TitanSpecial = True

def ApplyMountain(Player, Enemy):
    Player.TitanSpecial = False

def ApplyRuby(Player, Enemy):
    Player.IfritStack -= 1

def ApplyCyclone(Player, Enemy):
    Player.IfritSpecial = False
    Player.IfritSpecialCombo = True
    Player.EffectList.append(IfritCombo)

def ApplyStrike(Player, Enemy):
    Player.IfritSpecialCombo = False

def ApplyEmerald(Player, Enemy):
    Player.GarudaStack -= 1

def ApplySlipstream(Player, Enemy):
    Player.GarudaSpecial = False
    Player.SlipstreamDOTTimer = 15
    if Player.SlipstreamDOT == None : 
        Player.SlipstreamDOT = copy.deepcopy(SlipstreamDOT)
        Player.DOTList.append(Player.SlipstreamDOT)
        Player.EffectCDList.append(SlipstreamDOTCheck)
    else : Player.SlipstreamDOT.resetBuffSnapshot()

def ApplySummon(Player, Enemy):
    Player.SummonCD = 60 * Player.SpellReduction
    Player.Enkindle = True

    #Will check if we already have a summon object
    if Player.Pet == None:
        Pet(Player) # Summoning a pet for this player
        Player.Pet.ActionSet.append(WaitAbility(0.01)) #So program doesn't crash >.>
    else: Player.Pet.ResetStat() # Reseting stat on the summon

    # Can summon primals
    Player.TitanGem = True
    Player.GarudaGem = True
    Player.IfritGem = True



    Player.Pet.TrueLock = False #Delocking bahamut
    if not Player.LastTranceBahamut:
        #Then we summon bahamut
        Player.Deathflare = True
        Player.LastTranceBahamut = True
        Player.BahamutTrance = True
        Player.Pet.TrueLock = False
        Player.Pet.ActionSet.insert(Player.Pet.NextSpell + 1, BahamutAA) #Applying AA
    else:
        #Then we summon birdy
        Player.LastTranceBahamut = False
        Player.SummonDOT = copy.deepcopy(PhoenixDOT)
        Player.FirebirdTrance = True
        Player.Pet.TrueLock = False
        Player.Pet.ActionSet.insert(Player.Pet.NextSpell + 1, PhoenixAA)#Applying AA
    Player.EffectCDList.append(SummonDOTCheck)
    Player.SummonDOTTimer = 15

def ApplyEnkindle(Player, Enemy):
    Player.Enkindle = False
    Player.Pet.TrueLock = False
    Player.Pet.ActionSet.insert(Player.Pet.NextSpell + 1, EnkindleSummon)

def ApplyDeathflare(Player, Enemy):
    Player.Deathflare = False

def ApplyEnergyDrain(Player, Enemy):
    Player.AetherflowGauge = 2
    Player.AetherflowCD = 60
    Player.FurtherRuin = True

def ApplyFester(Player, Enemy):
    Player.AetherflowGauge -= 1

def ApplySearingLight(Player, Enemy):
    Player.SearingLightTimer = 30
    Player.SearingLightCD = 120
    Enemy.buffList.append(SearingLightbuff)
    Player.EffectCDList.append(SearingLightCheck)

def ApplyRekindle(Player, Enemy):
    pass

                                     # Only doing this if SavePreBakedAction is true
    #if Player.CurrentFight.SavePreBakedAction:
    #    fight = Player.CurrentFight
    #    history = buffPercentHistory(fight.TimeStamp, fight.TimeStamp + 30 , SearingLightbuff.MultDPS)
    #    fight.PlayerList[fight.PlayerIDSavePreBakedAction].PercentBuffHistory.append(history)


def ApplyBahamutAA(Player, Enemy):
    Player.DOTList.append(copy.deepcopy(BahamutDOT))

def ApplyPhoenixAA(Player, Enemy):
    Player.DOTList.append(copy.deepcopy(PhoenixDOT))

#Effect

def IfritCombo(Player, Spell):
    #Simply to make sure we loose the combo if we do another GCD
    if Spell.GCD:
        if Spell.id != Strike.id:
            Player.IfritSpecialCombo = False
        Player.EffectToRemove.append(IfritCombo)

#Check

def SlipstreamDOTCheck(Player, Enemy):
    if Player.SlipstreamDOTTimer <= 0:
        Player.DOTList.remove(Player.SlipstreamDOT)
        Player.EffectToRemove.append(SlipstreamDOTCheck)
        Player.SlipstreamDOT = None

def SummonDOTCheck(Player, Enemy):
    if Player.SummonDOTTimer <= 0:

        Player.Pet.DOTList = [] #Reseting DOTList

        Player.EffectToRemove.append(SummonDOTCheck)
        Player.FirebirdTrance = False
        Player.BahamutTrance = False
        Player.Deathflare = False

def SearingLightCheck(Player, Enemy):
    if Player.SearingLightTimer <= 0:
        Enemy.buffList.remove(SearingLightbuff)
        Player.EffectToRemove.append(SearingLightCheck)


#GCD
Ruin3 = SummonerSpell(3579, True, 1.5, 2.5, 310, 300, empty, [ManaRequirement], type = 1)
Ruin4 = SummonerSpell(7426, True, Lock, 2.5, 430, 400, ApplyRuin4, [ManaRequirement, Ruin4Requirement], type = 1)
FountainOfFire = SummonerSpell(16514, True, Lock, 2.5, 540, 300, empty, [FoFRequirement, ManaRequirement], type = 1)
BrandOfPurgatory = SummonerSpell(16515, True, Lock, 2.5, 240, 300, empty, [FoFRequirement, ManaRequirement], type = 1) # AOE of FoF
AstralImpulse = SummonerSpell(25820, True, Lock, 2.5, 440, 300, empty, [AstralImpulseRequirement, ManaRequirement], type = 1)
AstralFlare = SummonerSpell(25821, True, Lock, 2.5, 180, 300, empty, [AstralImpulseRequirement, ManaRequirement], type = 1)
TryDisaster = SummonerSpell(25826, True, 1.5, 2.5, 120, 300, empty, [ManaRequirement], type = 1)
#Primal
Titan = SummonerSpell(25839, True, Lock, 2.5, 750, 0, ApplyTitan, [TitanRequirement], type = 1)
Garuda = SummonerSpell(25840, True, Lock, 2.5, 750, 0, ApplyGaruda, [GarudaRequirement], type = 1)
Ifrit = SummonerSpell(25838, True, Lock, 2.5, 750, 0, ApplyIfrit, [IfritRequirement], type = 1)

#Titan Ability
Topaz = SummonerSpell(25824, True, Lock, 2.5, 330, 0, ApplyTopaz, [ManaRequirement, TopazRequirement], type = 1)
Mountain = SummonerSpell(25836, False, Lock, 0, 150, 0, ApplyMountain, [MountainRequirement])
TopazCatastrophe = SummonerSpell(25833, True, 0, 2.5, 140, 300, ApplyTopaz, [ManaRequirement, TopazRequirement] , type = 1) #same effect and requirement as Topaz

#IfritAbility
Ruby = SummonerSpell(25823, True, 2.8, 3, 510, 300, ApplyRuby, [RubyRequirement, ManaRequirement] , type = 1)
Cyclone = SummonerSpell(25835, True, Lock, 2.5, 430, 0, ApplyCyclone, [CycloneRequirement], type = 1)
Strike = SummonerSpell(25885, True, Lock, 2.5, 430, 0, ApplyStrike, [StrikeRequirement], type = 1)
RubyCatastrophe = SummonerSpell(25832, True, 2.8, 2.5, 210, 300, ApplyRuby, [RubyRequirement, ManaRequirement], type = 1) #Same as Ruby
#GarudaAbility
Emerald = SummonerSpell(25825, True, Lock, 1.5, 230, 300, ApplyEmerald, [EmeraldRequirement, ManaRequirement], type = 1)
EmeraldCatastrophe = SummonerSpell(25834, True, 0, 1.5, 100, 300, ApplyEmerald, [ManaRequirement, EmeraldRequirement], type = 1)
Slipstream = SummonerSpell(25837, True, 3, 3.5, 430, 0, ApplySlipstream, [SlipstreamRequirement], type = 1)
SlipstreamDOT = DOTSpell(100, 30, False, isGround=True)

#Summon
Summon = SummonerSpell(7427, True, Lock, 2.5, 0, 0, ApplySummon, [SummonRequirement], type = 1)
SummonPhoenix = SummonerSpell(25831, True, Lock, 2.5, 0, 0, ApplySummon, [SummonRequirement], type = 1)
#Bahamut and Phoenix damage will simply be a dot
BahamutAA = SummonerSpell(-115, False, 0, 0, 0, 0, ApplyBahamutAA, [])
PhoenixAA = SummonerSpell(-115, False, 0, 0, 0, 0, ApplyPhoenixAA, [])
BahamutDOT = DOTSpell(-14, 180, False) #AA of bahamut and phoenix
PhoenixDOT = DOTSpell(-15, 240, False)
EnkindleSummon = SummonerSpell(-17, False, 0, 0, 1300, 0, empty, []) #Enkindle done by pet
#autos of summon seems to be faster if uses Enkindle, but always max 5

#oGCD
Enkindle = SummonerSpell(7429, False, 0, 0, 0, 0, ApplyEnkindle, [EnkindleRequirement]) #Smaller lock since executed by pet, might have to reconsider... >.>
EnkindlePhoenix = SummonerSpell(16516, False, 0, 0, 0, 0, ApplyEnkindle, [EnkindleRequirement])
Deathflare = SummonerSpell(3582, False, Lock, 0, 500, 0, ApplyDeathflare, [DeathflareRequirement])
Rekindle = SummonerSpell(25830, False, 0, 0 ,0, 0, ApplyRekindle, [RekindleRequirement])
EnergyDrainSMN = SummonerSpell(16508, False, Lock, 0, 200, 0, ApplyEnergyDrain, [EnergyDrainRequirement])
Fester = SummonerSpell(181, False, Lock, 0, 340, 0, ApplyFester, [FesterRequirement])
PainFlare = SummonerSpell(3578, False, 0, 0, 150, 0, ApplyFester, [FesterRequirement]) #AOE version of Fester
SearingLight = SummonerSpell(25801, False, Lock, 0, 0, 0, ApplySearingLight, [SearingLightRequirement])
PhysickSMN = SummonerSpell(16230, True, 1.5, 2.5, 0, 0, empty, [ManaRequirement])
Resurrection = SummonerSpell(173, True, 8, 2.5, 0, 2400, empty, [ManaRequirement])
RadiantAegis = SummonerSpell(25799, True, 0, 0, 0, 0, empty, [])
#buff
SearingLightbuff = buff(1.03,name="SL")

SummonerAbility = {
3579 : Ruin3,
7426 : Ruin4,
25826 : TryDisaster,
25801 : SearingLight,
#25789 : Carbunle,
7427 : Summon, #Bahamut
25820 : AstralImpulse,
25821 : AstralFlare, 
3582 : Deathflare,
7429 : Enkindle, #Enkindle Bahamut
#7428 : Wyrmwave, #Bahamut Autos
25831 : SummonPhoenix, #Summon Phoenix
16514 : FountainOfFire, #Fountain Of Fire
16515 : BrandOfPurgatory, #AOE of FoF
16516 : EnkindlePhoenix, #Phoenix Enkindle
25830 : Rekindle,
#16519 : ScarletFlame, #Pheonix AA
25838 : Ifrit,
25823 : Ruby,
25832 : RubyCatastrophe,
25835 : Cyclone,
25885 : Strike,
25839 : Titan,
25824 : Topaz,
25833 : TopazCatastrophe,
25836 : Mountain,
25840 : Garuda,
25825 : Emerald,
25834 : EmeraldCatastrophe,
25837 : Slipstream,
16508 : EnergyDrainSMN,
16510 : EnergyDrainSMN, #Energy Syphon
181 : Fester,
3578 : PainFlare,
25799 : RadiantAegis,
16230 : PhysickSMN,
173 : Resurrection
}


"""
This module is outdated and should not be used. An error is thrown when importing it but if you still want to use it
simply do a try/except block.
"""

from ffxivcalc.helperCode.exceptions import outDatedAPICode
raise outDatedAPICode



"""
This file contains the parser will translate code in a way the simulator can use.


The language contains the following :

SET fieldName Value -> This will set the given fieldName to a given value. Type checking is done on the value that was previously there
                                                                           and an error will be thrown if type does not match.
FORCESET fieldName Value -> Same as SET, but does not do type check and forces a new value on the field.

Every command must be delimited by a ; . Invalid text within ; will be ignored.
"""

from copy import deepcopy

commandList = ["SET", "FORCESET"]

commandExpectedInputAmount = {"SET" : 2, "FORCESET" : 2}

class parser:

    def __init__(self):
        pass

    def cleanCommandComponent(self, commandComponent : list) -> bool:
        """
        This command cleans up the given commandComponent and returns True if the command is valid.
        """
                             # Removing empty parts that might have been caused by missinputs
        commandComponent = self.filterEmpty(commandComponent)

                             # Check if initial entry is a valid command in commandList
        if not (commandComponent[0] in commandList) : return [],False

        return commandComponent, len(commandComponent[1:]) == commandExpectedInputAmount[commandComponent[0]]

    def filterEmpty(self, input):
        """
        This command filters the empty index of a list.
        """

        def emptyInput(s : str) -> bool:
            return len(s) != 0

        newInput = []

        for x in filter(emptyInput, input): newInput.append(x)

        return newInput

    def parseString(self, code : str) -> dict:
        """
        This function returns a dictionnary with the name of every fields as key and the value we want to set
        this field as. This does not type check and simply reads and output.
        """
                             # This list contains all command formatted in a way the simulator can understand.
        commandListCompiled = []

                             # Seperating by semi-colons (;)
        commandList = code.split(";")
        
        commandList = self.filterEmpty(commandList)

        for command in commandList:
            commandComponent = command.split(" ")
            commandComponent, valid = self.cleanCommandComponent(commandComponent)
            if not valid: continue

                             # This is a list containing all information of the command in order.
            commandListFormat = []

            for i in range(commandExpectedInputAmount[commandComponent[0]]):
                commandListFormat.append(commandComponent[i+1])

            commandListCompiled.append({commandComponent[0] : deepcopy(commandListFormat)})

        return commandListCompiled
    

mainParser = parser()

def executeCode(sourceCode : str, targetObject):
    """
    This function receives as input sourceCode in the style defined above and
    applies the code to the targetobject.
    """

    commandList = mainParser.parseString(sourceCode)

    possibleFieldName = targetObject.__dict__.keys()

    for command in commandList:

        commandName = list(command.keys())[0]

        match commandName:
            case "FORCESET":
                fieldName = command["FORCESET"][0]
                if fieldName in possibleFieldName: 
                    targetObject.__dict__[fieldName] = command["FORCESET"][1]
            case "SET":
                fieldName = command["SET"][0]
                if fieldName in possibleFieldName: 
                    match targetObject.__dict__[fieldName]:
                        case int() : targetObject.__dict__[fieldName] = int(command["SET"][1])
                        case float() : targetObject.__dict__[fieldName] = float(command["SET"][1])
                        case str() : targetObject.__dict__[fieldName] = str(command["SET"][1])
                        case bool() : targetObject.__dict__[fieldName] = bool(command["SET"][1])




    


from ffxivcalc.Jobs.Base_Spell import DOTSpell, ManaRequirement, buff, empty
from ffxivcalc.Jobs.Caster.Caster_Spell import BLMSpell, SwiftcastEffect
from ffxivcalc.helperCode.helper_math import roundDown
import copy
import math
Lock = 0
#Requirement

def ManawardRequirement(Player, Spell):
    return Player.ManawardCD <= 0, Player.ManawardCD

def EnochianRequirement(Player, Spell):
    return Player.EnochianTimer > 0, -1

def FireRequirement(Player, Spell):
    return Player.ElementalGauge > 0, -1

def IceRequirement(Player, Spell):
    return Player.ElementalGauge < 0, -1

def ParadoxRequirement(Player, Spell):
    return Player.Paradox, -1

def PolyglotRequirement(Player, Spell):
    return Player.PolyglotStack > 0, Player.PolyglotTimer

def TransposeRequirement(Player, Spell):
    return Player.TransposeCD <= 0, Player.TransposeCD

def AmplifierRequirement(Player, Spell):
    return Player.AmplifierCD <= 0, Player.AmplifierCD

def LeyLinesRequirement(Player, Spell):
    return Player.LeyLinesCD <= 0, Player.LeyLinesCD

def TripleCastRequirement(Player, Spell):
    return Player.TripleCastUseStack > 0, Player.TripleCastCD

def SharpCastRequirement(Player, Spell):
    return Player.SharpCastStack > 0, Player.SharpCastCD

def ManafrontRequirement(Player, Spell):
    return Player.ManafrontCD <= 0, Player.ManafrontCD

#Apply

def ApplyManaward(Player, Enemy):
    Player.ManawardCD = 120

def ApplyFlare(Player, Enemy):
    manaback = 0
    #input(Player.UmbralHearts)
    if Player.UmbralHearts > 0:
        Player.UmbralHearts -= 1
        manaback = math.floor(Player.Mana/3)
    ApplyDespair(Player, Enemy)
    Player.Mana += manaback #we get 1/3rd of our mana back if with UmbralHearts stack

def ApplyHighFire(Player, Enemy):
    ApplyFire3(Player, Enemy)
    if not HighFireEffect in Player.EffectList:
        Player.EffectList.insert(0,HighFireEffect)
        Player.EffectCDList.append(HighFireCheck)

def ApplyThunder4(Player, Enemy):
    if Player.Thunder4DOT == None:
        Player.Thunder4DOT = copy.deepcopy(Thunder4DOT)
        Player.DOTList.append(Player.Thunder4DOT)
        Player.EffectCDList.append(Thunder4DOTCheck)
    else: Player.Thunder4DOT.resetBuffSnapshot() # If already applied reset snapshot
    Player.Thunder4DOTTimer = 18

    if Player.SharpCast: #If we have SharpCast
        Player.EffectList.insert(0,Thunder3ProcEffect)
        Player.SharpCast = False

    #We now have to check if Thunder 3 is already applied, in which case we will remove it
    if Player.Thunder3DOT != None : Player.Thunder3DOTTimer = 0

def ApplyUmbralSoul(Player, Enemy):
    Player.AddIce() #Add 1 Ice
    Player.UmbralHearts = min(3, Player.UmbralHearts + 1)

def ApplyBlizzard1(Player, Enemy):
    Player.AddIce() #Add 1 Ice

def ApplyFire1(Player, Enemy):
    Player.AddFire() #Add 1 fire
    if Player.SharpCast: #If sharpcast
        Player.Fire3Proc = True
        Player.SharpCast = False
        Player.EffectList.insert(0,Fire3ProcEffect)

def ApplyFire3(Player, Enemy):
    #Will check if we unlock paradox
    if Player.ElementalGauge == -3 and Player.UmbralHearts == 3:
        Player.Paradox = True

    Player.ElementalGauge = 3
    Player.EnochianTimer = 15 #Reset Timer

def ApplyBlizzard3(Player, Enemy):

    #Check if we unlock paradox
    if Player.ElementalGauge == 3: Player.Paradox = True

    Player.ElementalGauge = -3
    Player.EnochianTimer = 15 #ResetTimer

def ApplyBlizzard4(Player, Enemy):
    Player.UmbralHearts = 3

def ApplyParadox(Player, Enemy):
    Player.Paradox = False

    if Player.ElementalGauge > 0:
        Player.AddFire()
        if Player.SharpCast: #If sharpcast
            Player.Fire3Proc = True
            Player.SharpCast = False
            Player.EffectList.insert(0,Fire3ProcEffect)
    elif Player.ElementalGauge < 0:
        Player.AddIce()

def ApplyXenoglossy(Player, Spell):
    Player.PolyglotStack -= 1

def ApplyDespair(Player, Spell):
    Player.Mana = 0 #All mana is used
    Player.EnochianTimer = 15
    Player.ElementalGauge = 3

def ApplyThunder3(Player, Spell):

    if Player.Thunder3DOT == None: #If no dot already applied
        Player.Thunder3DOT = copy.deepcopy(Thunder3DOT)
        Player.DOTList.append(Player.Thunder3DOT)
        Player.EffectCDList.append(Thunder3DOTCheck)
    else: Player.Thunder3DOT.resetBuffSnapshot() # If already applied reset snapshot
    Player.Thunder3DOTTimer = 30

    if Player.SharpCast: #If we have SharpCast
        Player.EffectList.insert(0,Thunder3ProcEffect)
        Player.SharpCast = False

    #We now have to check if Thunder 4 is already applied, in which case we will remove it
    if Player.Thunder4DOT != None : Player.Thunder4DOTTimer = 0

def ApplyTranspose(Player, Spell):
    Player.TransposeCD = 4
    #Check if we unlock paradox

    if (Player.ElementalGauge == 3) or (Player.ElementalGauge == -3 and Player.UmbralHearts == 3): Player.Paradox = True #In fire phase with 3 Astral Fire

    if Player.ElementalGauge > 0: Player.ElementalGauge = -1 #Ice
    elif Player.ElementalGauge < 0 : Player.ElementalGauge = 1 #Fire

def ApplyAmplifier(Player, Enemy):
    Player.AmplifierCD = 120
    Player.PolyglotStack = min(2, Player.PolyglotStack + 1)

def ApplyLeyLines(Player, Enemy):
    Player.LeyLinesCD = 120
    Player.LeyLinesTimer = 30

    Player.EffectList.append(LeyLinesEffect)
    Player.EffectCDList.append(LeyLinesCheck)

    Player.Haste += 15
    Player.hasteHasChanged = True
    Player.hasteChangeValue = 15

def ApplyTripleCast(Player, Enemy):
    #Check if we need to add stack check
    if Player.TripleCastUseStack == 2:
        Player.EffectCDList.append(TripleCastUseStackCheck)
        Player.TripleCastCD = 60
    Player.TripleCastUseStack -= 1

    if Player.TripleCastStack == 0: #If not stack of insta-cast
        Player.EffectList.append(TripleCastEffect)
        Player.EffectCDList.append(TripleCastCheck)
    Player.TripleCastStack = 3

def ApplySharpCast(Player, Enemy):
    #Check if we needto add stack check
    if Player.SharpCastStack == 2:
        Player.EffectCDList.append(SharpCastStackCheck)
        Player.SharpCastCD = 30
    Player.SharpCastStack -= 1

    Player.SharpCast = True

def ApplyManafront(Player, Enemy):
    Player.ManafrontCD = 120
    Player.Mana = min(10000, Player.Mana + 3000)

#Effect

def HighFireEffect(Player, Spell):
    if Spell.id == Flare.id:
        Spell.Potency += 60
        Player.EffectToRemove.append(HighFireEffect)
        Player.EffectCDList.remove(HighFireCheck)

def Fire3ProcEffect(Player, Spell):
    if Spell.id == Fire3.id:
        Spell.ManaCost = 0
        Spell.CastTime = Lock
        Player.EffectToRemove.append(Fire3ProcEffect)

def Thunder3ProcEffect(Player, Spell):
    if Spell.id == Thunder3.id:
        Spell.Potency += 350
        Spell.ManaCost = 0
        Spell.CastTime = Lock
        Player.EffectToRemove.append(Thunder3ProcEffect)
    elif Spell.id == Thunder4.id:
        Spell.Potency += 120
        Spell.ManaCost = 0
        Spell.CastTime = Lock
        Player.EffectToRemove.append(Thunder3ProcEffect)

def TripleCastEffect(Player, Spell): 
    if not (SwiftcastEffect in Player.EffectList) and Spell.GCD and Spell.CastTime > 0 : #If GCD and not already insta cast, also Swift will go before
        Spell.CastTime = Lock
        Player.TripleCastStack -= 1

def LeyLinesEffect(Player, Spell):
    if Spell.GCD:
        pass
        #Spell.CastTime = (Spell.CastTime * 0.85)
        #Spell.RecastTime = (Spell.RecastTime * 0.85)

#def EnochianEffect(Player, Spell):
#    if Player.ElementalGauge != 0:
#        #If elementalGauge is not 0
#        Player.buffList.append(Enochian)
#        Player.EffectToRemove.append(EnochianEffect)
#        Player.EffectCDList.append(EnochianEffectCheck)
#        Player.Enochian = True

def ElementalEffect(Player, Spell):
    #Will affect Spell depending on fire and ice
    #input("UmbrealHearts : " + str(Player.UmbralHearts))
    if isinstance(Spell, BLMSpell) and Spell.IsFire:
        #Fire Spell
        #First check if fire phase and apply Effect
        if Player.ElementalGauge > 0: #Fire Phase
            #input('we in : ' + str(Spell.id))
            if Player.ElementalGauge == 1: Spell.Potency *= 1.4
            elif Player.ElementalGauge == 2 : Spell.Potency *= 1.6
            elif Player.ElementalGauge == 3 : Spell.Potency *= 1.8

            if (Player.UmbralHearts > 0) and Spell.id!= Flare.id: #If we have UmbralHearts, then no mana cost increase. We also have to make sure
                #it isn't Flare, since Flare has its own way of dealing with UmbrealHearts
                Player.UmbralHearts -= 1
            else : 
                if Spell.id != Despair.id and Spell.id != Flare.id: Spell.ManaCost *= 2 #Double mana cost, only if not despair
        #Check if ice phase
        elif Player.ElementalGauge < 0 : #Ice Phase
            if Player.ElementalGauge == -1: 
                Spell.Potency *= 0.9
                Spell.ManaCost *= 0.75
            elif Player.ElementalGauge == -2 : 
                Spell.Potency *= 0.8
                Spell.ManaCost *= 0.5
            elif Player.ElementalGauge == -3 : 
                Spell.Potency *= 0.7
                Spell.ManaCost = 0
                Spell.CastTime *= 0.5

    elif isinstance(Spell, BLMSpell) and Spell.IsIce:
        #Ice Spell
        #First check if fire phase
        if Player.ElementalGauge > 0: #Fire Phase
            Spell.ManaCost = 0
            if Player.ElementalGauge == 1: Spell.Potency *= 0.9
            elif Player.ElementalGauge == 2 : Spell.Potency *= 0.8
            elif Player.ElementalGauge == 3 : 
                Spell.Potency *= 0.7
                Spell.CastTime *= 0.5
        elif Player.ElementalGauge < 0 : #Ice Phase
            if Player.ElementalGauge == -1: Spell.ManaCost *= 0.75
            elif Player.ElementalGauge == -2 : Spell.ManaCost *= 0.5
            elif Player.ElementalGauge == -3 : Spell.ManaCost = 0

    elif Spell.id == Paradox.id:
        #If we are casting Paradox

        if Player.ElementalGauge < 0 : #Ice Phase
            Spell.ManaCost = 0
            Spell.CastTime = 0

#Check

def HighFireCheck(Player, Enemy):
    if Player.ElementalGauge <= 0: #If loose astralFire
        Player.EffectList.remove(HighFireEffect)
        Player.EffectToRemove.append(HighFireCheck)

def EnochianEffectCheck(Player, Enemy):

    if Player.EnochianTimer <= 0 : Player.ElementalGauge = 0

    if Player.ElementalGauge != 0 and not Player.Enochian:
        # Receive enochian
        Player.Enochian = True
        Player.buffList.append(Enochian)
        Player.PolyglotTimer = 30

    if Player.ElementalGauge == 0 and Player.Enochian : #If we loose Enochian
        Player.Enochian = False
        Player.buffList.remove(Enochian)
        Player.PolyglotTimer = 30

    if Player.PolyglotTimer <= 0 and Player.Enochian:
        #Add new stack
        Player.PolyglotStack = min(2, Player.PolyglotStack + 1)
        Player.PolyglotTimer = 30


def SharpCastStackCheck(Player, Enemy):
    if Player.SharpCastCD <= 0:
        if Player.SharpCastStack == 1:
            Player.EffectToRemove.append(SharpCastStackCheck)
        else:
            Player.SharpCastCD = 30
        Player.SharpCastStack += 1

def TripleCastUseStackCheck(Player, Enemy):
    if Player.TripleCastCD <= 0:
        if Player.TripleCastUseStack == 1:
            Player.EffectToRemove.append(TripleCastUseStackCheck)
        else:
            Player.TripleCastCD = 60
        Player.TripleCastUseStack += 1

def TripleCastCheck(Player, Enemy):
    if Player.TripleCastStack == 0 :
        Player.EffectList.remove(TripleCastEffect)
        Player.EffectToRemove.append(TripleCastCheck)

def LeyLinesCheck(Player, Enemy):
    if Player.LeyLinesTimer <= 0:
        Player.EffectList.remove(LeyLinesEffect)
        Player.Haste -= 15
        Player.hasteChangeValue = -15
        Player.hasteHasChanged = True
        Player.EffectToRemove.append(LeyLinesCheck)

def Thunder3DOTCheck(Player, Enemy):
    if Player.Thunder3DOTTimer <= 0:
        Player.DOTList.remove(Player.Thunder3DOT)
        Player.Thunder3DOT = None
        Player.EffectToRemove.append(Thunder3DOTCheck)

def Thunder4DOTCheck(Player, Enemy):
    if Player.Thunder4DOTTimer <= 0:
        Player.DOTList.remove(Player.Thunder4DOT)
        Player.Thunder4DOT = None
        Player.EffectToRemove.append(Thunder4DOTCheck)

#GCD

def ApplyFire4(Player, Enemy):
    input("Applying fire4 at : " + str(Player.CurrentFight.TimeStamp))

#Fire Spell
Fire1 = BLMSpell(141, True, 2.5, 2.5, 180, 800, True, False, ApplyFire1, [ManaRequirement], type = 1)
Fire2 = BLMSpell(147, True, 3, 2.5, 100, 1500, True, False, ApplyFire3, [ManaRequirement], type = 1) #Same effect as Fire 3
Fire3 = BLMSpell(152, True, 3.5, 2.5, 260, 2000, True, False, ApplyFire3, [ManaRequirement], type = 1)
Fire4 = BLMSpell(3577, True, 2.8, 2.5, 310, 800, True, False, empty, [EnochianRequirement, FireRequirement, ManaRequirement], type = 1) #BIG PP DAMAGE LETS GOOOOOOOOOOOOOo
Despair = BLMSpell(16505, True, 3, 2.5, 340, 800, True, False, ApplyDespair, [FireRequirement, ManaRequirement], type = 1)
Flare = BLMSpell(162, True, 4, 2.5, 220, 800, True, False, ApplyFlare, [ManaRequirement, FireRequirement], type = 1)
HighFire = BLMSpell(25794, True, 3, 2.5, 140, 1500, True, False, ApplyHighFire, [ManaRequirement], type = 1)



#Ice Spell
UmbralSoul = BLMSpell(16506, True, 0, 2.5, 0, 0, False, True, ApplyUmbralSoul, [IceRequirement], type = 1)
Blizzard1 = BLMSpell(142, True, 2.5, 2.5, 180, 400, False, True, ApplyBlizzard1, [ManaRequirement], type = 1)
Blizzard3 = BLMSpell(154, True, 3.5, 2.5, 260, 800, False, True, ApplyBlizzard3, [ManaRequirement], type = 1)
Blizzard4 = BLMSpell(3576, True, 2.5, 2.5, 310, 800, False, True, ApplyBlizzard4, [EnochianRequirement, IceRequirement, ManaRequirement], type = 1)
Freeze = BLMSpell(159, True, 2.8, 2.5, 120, 1000, False, True, ApplyBlizzard4, [EnochianRequirement, IceRequirement, ManaRequirement], type = 1) #Same as B4
HighBlizzard = BLMSpell(25795, True, 3, 2.5, 140, 800, False, True, ApplyBlizzard3, [ManaRequirement], type = 1)

#Unaspected Spell
Scathe = BLMSpell(156, True, Lock, 2.5, 100, 800, False, False, empty, [ManaRequirement], type = 1)
Paradox = BLMSpell(25797, True, 2.5, 2.5, 500, 1600, False, False, ApplyParadox, [ParadoxRequirement, ManaRequirement], type = 1) 
Xenoglossy = BLMSpell(16507, True, Lock, 2.5, 880, 0, False, False, ApplyXenoglossy, [PolyglotRequirement], type = 1)
Foul = BLMSpell(7422, True, Lock, 2.5, 600, 0, False, False, ApplyXenoglossy, [PolyglotRequirement], type = 1) #Same effect as Xeno
Thunder3 = BLMSpell(153, True, 2.5, 2.5, 50, 400, False, False, ApplyThunder3, [ManaRequirement], type = 1)
Thunder3DOT = DOTSpell(-21, 35, False)
Thunder4 = BLMSpell(7420, True, 2.5, 2.5, 50, 400, False, False, ApplyThunder4, [ManaRequirement], type = 1)
Thunder4DOT = DOTSpell(-40, 20, False)

#oGCD
Transpose = BLMSpell(149, False, Lock, 0, 0, 0, False, False, ApplyTranspose, [TransposeRequirement])
Amplifier = BLMSpell(25796, False, Lock, 0, 0, 0, False, False, ApplyAmplifier, [AmplifierRequirement])
LeyLines = BLMSpell(3573, False, Lock, 0, 0, 0, False, False, ApplyLeyLines, [LeyLinesRequirement])
Triplecast = BLMSpell(7421, False, Lock, 0, 0, 0, False, False, ApplyTripleCast, [TripleCastRequirement])
SharpCast = BLMSpell(3574, False, Lock, 0, 0, 0, False, False, ApplySharpCast, [SharpCastRequirement])
Manafront = BLMSpell(158, False, Lock, 0, 0, 0, False, False, ApplyManafront, [ManafrontRequirement])
BetweenTheLine = BLMSpell(7419, False, 0, 0, 0, 0, False, False, empty, [])
AetherialManipulation = BLMSpell(155, False, 0, 0, 0, 0, False, False, empty, [])
Manaward = BLMSpell(157, False, 0, 0, 0, 0, False, False, ApplyManaward, [ManawardRequirement])

#buff
Enochian = buff(1.23, name="Enochian")


#All BlackMage abilities with their id in a dictionnary

BlackMageAbility = {
141 : Fire1, 
142 : Blizzard1, 
147 : Fire2, 
149 : Transpose, 
152 : Fire3,
155 : AetherialManipulation, 
156 : Scathe,
157 : Manaward, 
153 : Thunder3, 
154 : Blizzard3, 
158 : Manafront, 
159 : Freeze, 
162 : Flare, 
3573 : LeyLines, 
3574 : SharpCast, 
3576 : Blizzard4, 
3577 : Fire4, 
7419 : BetweenTheLine, 
7420 : Thunder4, 
7421 : Triplecast, 
7422 : Foul, 
16505 : Despair, 
16506 : UmbralSoul, 
16507 : Xenoglossy, 
25794 : HighFire, 
25795 : HighBlizzard, 
25796 : Amplifier, 
25797 : Paradox
}


from ffxivcalc.Jobs.Base_Spell import ManaRequirement, Potion, Spell, empty
from ffxivcalc.Jobs.Caster.Caster_Spell import LucidDreaming, Surecast, Swiftcast
Lock = 0.75
class HealerSpell(Spell):

    def __init__(self, id, GCD, CastTime, RecastTime, Potency, ManaCost, Effect, Requirement, type = 0, AOEHeal = False, TargetHeal = False):
        super().__init__(id, GCD, CastTime, RecastTime, Potency, ManaCost, Effect, Requirement, type = type, AOEHeal = AOEHeal, TargetHeal=TargetHeal)


#########################################
########## SCHOLAR SPELL  ###############
#########################################

class ScholarSpell(HealerSpell):

    def __init__(self, id, GCD, CastTime, RecastTime, Potency, ManaCost, Effect, Requirement, type = 0, AOEHeal = False, TargetHeal = False):
        super().__init__(id, GCD, CastTime, RecastTime, Potency, ManaCost, Effect, Requirement, type = type, AOEHeal = AOEHeal, TargetHeal=TargetHeal)


#########################################
########## WHITEMAGE SPELL ##############
#########################################


class WhitemageSpell(HealerSpell):
    def __init__(self, id, GCD, CastTime, RecastTime, Potency, ManaCost, Effect, Requirement, type = 0, AOEHeal = False, TargetHeal = False):
        super().__init__(id, GCD, CastTime, RecastTime, Potency, ManaCost, Effect, Requirement, type = type, AOEHeal = AOEHeal, TargetHeal=TargetHeal)

#########################################
########## ASTROLOGIAN SPELL ############
#########################################


class AstrologianSpell(HealerSpell):
    def __init__(self, id, GCD, CastTime, RecastTime, Potency, ManaCost, Effect, Requirement, type = 0, AOEHeal = False, TargetHeal = False):
        super().__init__(id, GCD, CastTime, RecastTime, Potency, ManaCost, Effect, Requirement, type = type, AOEHeal = AOEHeal, TargetHeal=TargetHeal)

#########################################
##########     SAGE   SPELL   ###########
#########################################


class SageSpell(HealerSpell):
    def __init__(self, id, GCD, CastTime, RecastTime, Potency, ManaCost, Effect, Requirement, type = 0, AOEHeal = False, TargetHeal = False):
        super().__init__(id, GCD, CastTime, RecastTime, Potency, ManaCost, Effect, Requirement, type = type, AOEHeal = AOEHeal, TargetHeal=TargetHeal)



#Class Action spell

#Requirement
def RescueRequirement(Player, Spell):
    return Player.RescueCD <= 0, Player.RescueCD

#Apply

def ApplyRescue(Player, Enemy):
    Player.RescueCD = 120

#Swiftcast, Surecast and LucidDreaming are in Caster_Spell.py
Repose = HealerSpell(16560, True, 2.5, 2.5, 0, 600, empty, [ManaRequirement])
Esuna = HealerSpell(7568, True, 1, 2.5, 0, 400, empty, [ManaRequirement])
Rescue = HealerSpell(7571, False, 0, 0, 0, 0, ApplyRescue, [RescueRequirement])

# Limit Break actions

LB1Timer = 4.10
LB2Timer = 7.13
LB3Timer = 10.10

HealerLB1 = HealerSpell(1111, False,LB1Timer, LB1Timer, 0, 0, empty, [], type=3)
HealerLB2 = HealerSpell(1112, False,LB2Timer, LB2Timer, 0, 0, empty, [], type=3)
HealerLB3 = HealerSpell(1113, False,LB3Timer, LB3Timer, 0, 0, empty, [], type=3)

HealerAbility = {
16560 : Repose,
7568 : Esuna,
7571 : Rescue,
7561 : Swiftcast,
7562 : LucidDreaming,
7559 : Surecast,
34590545 : Potion,
34592399 : Potion, 
34594163 : Potion,
-2 : Potion,
1111 : HealerLB1,
1112 : HealerLB2,
1113 : HealerLB3
}


#########################################
########## REDMAGE PLAYER ###############
#########################################
# Acc -> Dual -> Swift
from ffxivcalc.Jobs.Base_Spell import ManaRequirement, buff, empty, buffPercentHistory
from ffxivcalc.Jobs.Caster.Caster_Spell import RedmageSpell, SwiftcastEffect
Lock = 0.75
#Special

def Addmana(Player, WhiteMana, BlackMana):
    Player.WhiteMana = min(100, Player.WhiteMana + WhiteMana)
    Player.BlackMana = min(100, Player.BlackMana + BlackMana)

def Removemana(Player, WhiteMana, BlackMana):
    Player.WhiteMana = max(0, Player.WhiteMana - WhiteMana)
    Player.BlackMana = max(0, Player.BlackMana - BlackMana)

#Requirement
def RDMManaRequirement(Player, Spell):
    Removemana(Player, Spell.WhiteCost, Spell.BlackCost)
    return Spell.ManaCost <= Player.Mana and Player.BlackMana >= 0 and Player.WhiteMana >= 0, -1

def ManaficationRequirement(Player, Spell):
    return Player.ManaficationCD <= 0, Player.ManaficationCD

def EmboldenRequirement(Player, Spell):
    return Player.EmboldenCD <= 0, Player.EmboldenCD

def AccelerationRequirement(Player, Spell):
    return Player.AccelerationStack >= 1, Player.AccelerationCD

def FlecheRequirement(Player, Spell):
    return Player.FlecheCD <= 0, Player.FlecheCD

def ContreRequirement(Player, Spell):
    return Player.ContreCD <= 0, Player.ContreCD

def EngagementRequirement(Player, Spell):
    return Player.EngagementStack > 0, Player.EngagementCD

def CorpsRequirement(Player, Spell):
    return Player.CorpsStack > 0, Player.CorpsCD

#Combo Action Requirements

def ZwerchhauRequirement(Player, Spell):
    return Player.Zwerchhau,-1

def RedoublementRequirement(Player, Spell):
    return Player.Redoublement,-1

def VerholyRequirement(Player, Spell):
    return Player.ManaStack == 3,-1

def ScorchRequirement(Player, Spell):
    return Player.Scorch,-1

def ResolutionRequirement(Player, Spell):
    return Player.Resolution,-1

def MagickBarrierRequirement(Player, Spell):
    return Player.MagickBarrierCD <= 0, Player.MagickBarrierCD

#Apply

def ApplyEnchantedMoulinet(Player, Enemy):
    Player.ManaStack = min(3, Player.ManaStack + 1) #Max of 3 stacks

def ApplyMagickBarrier(Player, Enemy):
    Player.MagickBarrierCD = 120

def ApplyImpact(Player, Enemy):
    Addmana(Player, 3, 3) #Add mana

def ApplyMoulinet(Player, Enemy):
    #Removemana(Player, 20, 20)
    Player.ManaStack = min(3, Player.ManaStack + 1) #Max of 3 stacks

def ApplyJolt(Player, Enemy):
    Addmana(Player, 2, 2)

def ApplyVerfire(Player, Enemy):
    Player.UsedVerfireProc += 1
    Addmana(Player, 0, 5)

def ApplyVerstone(Player,Enemy):
    Player.UsedVerstoneProc += 1
    Addmana(Player, 5, 0)

def ApplyVerareao(Player, Enemy):
    Player.ExpectedVerstoneProc += 0.5
    Addmana(Player, 6, 0)

def ApplyVerthunder(Player, Enemy):
    Player.ExpectedVerfireProc += 0.5
    Addmana(Player, 0, 6)

def ApplyManafication(Player, Enemy):
    Player.ManaficationStack = 6
    Player.ManaficationCD = 110
    Player.EffectList.insert(0,ManaficationEffect)
    Player.EffectCDList.append(ManaficationCheck)
    Addmana(Player, 50, 50)

def ApplyEmbolden(Player, Enemy):
    Player.EmboldenTimer = 20
    Player.EmboldenCD = 120
    Player.EffectCDList.append(EmboldenCheck)
    Enemy.buffList.append(EmboldenBuff) #5% DPS boost for everyone. I put buff on Boss, but could put it on every other player
        
                                     # Only doing this if SavePreBakedAction is true
    #if Player.CurrentFight.SavePreBakedAction:
    #    fight = Player.CurrentFight
    #    history = buffPercentHistory(fight.TimeStamp, fight.TimeStamp + 20 , EmboldenBuff.MultDPS)
    #    fight.PlayerList[fight.PlayerIDSavePreBakedAction].PercentBuffHistory.append(history)

def ApplyAcceleration(Player, Enemy):
    if Player.AccelerationStack == 2:
         Player.AccelerationCD = 55
         Player.EffectCDList.append(AccelerationStackCheck)
    Player.EffectList.insert(0,AccelerationEffect)
    Player.AccelerationStack -= 1

def ApplyFleche(Player, Enemy):
    Player.FlecheCD = 25

def ApplyContre(Player, Enemy):
    Player.ContreCD = 35

def ApplyEngagement(Player, Enemy):
    if Player.EngagementStack == 2:
        Player.EngagementCD = 35
        Player.EffectCDList.append(EngagementStackCheck)
    Player.EngagementStack -= 1

def ApplyCorps(Player, Enemy):
    if Player.CorpsStack == 2:
        Player.CorpsCD = 35
        Player.EffectCDList.append(CorpsStackCheck)
    Player.CorpsStack -= 1




#Combo Action Apply

def ApplyRiposte(Player, Enemy):
    if not (Riposte in Player.EffectList) : Player.EffectList.append(RiposteCombo)

def ApplyEnchantedRiposte(Player, Enemy):
    #Removemana(Player, 20, 20)
    Player.ManaStack = min(3, Player.ManaStack + 1) #Max of 3 stacks
    if not (Riposte in Player.EffectList) : Player.EffectList.append(RiposteCombo)
    if not (ManaStackEffect in Player.EffectList) : Player.EffectList.append(ManaStackEffect)
    #This effect is to make sure we only do melee actions, since otherwise we loose mana stacks

def ApplyZwerchhau(Player, Enemy):
    #Removemana(Player, 15, 15)
    Player.ManaStack = min(3, Player.ManaStack + 1) #Max of 3 stacks
    if not (ManaStackEffect in Player.EffectList) : Player.EffectList.append(ManaStackEffect)
    #This effect is to make sure we only do melee actions, since otherwise we loose mana stacks

def ApplyRedoublement(Player, Enemy):
    #Removemana(Player, 15, 15)
    Player.ManaStack = min(3, Player.ManaStack + 1) #Max of 3 stacks
    if not (ManaStackEffect in Player.EffectList) : Player.EffectList.append(ManaStackEffect)
    #This effect is to make sure we only do melee actions, since otherwise we loose mana stacks

def ApplyVerholy(Player, Enemy):
    Addmana(Player, 11, 0)
    Player.Scorch = True
    Player.ManaStack -= 1 #Removing one mana stack
    if Player.WhiteMana < Player.BlackMana:
        Player.ExpectedVerstoneProc += 1
    else : Player.ExpectedVerstoneProc += 0.2

def ApplyVerflare(Player, Enemy):
    Addmana(Player, 0, 11)
    Player.Scorch = True
    Player.ManaStack -= 1 #Removing one mana stack
    if Player.WhiteMana > Player.BlackMana:
        Player.ExpectedVerfireProc += 1
    else : Player.ExpectedVerfireProc += 0.2

def ApplyScorch(Player, Enemy):
    Addmana(Player, 4, 4)
    Player.Resolution = True
    Player.Scorch = False
    Player.ManaStack -= 1 #Removing one mana stack

def ApplyResolution(Player, Enemy):
    Addmana(Player, 4, 4)
    Player.Resolution = False
    Player.ManaStack -= 1 #Removing one mana stack


#Effect

def ManaStackEffect(Player, Spell):
    if Spell.id == Verholy.id or Spell.id == Verflare.id:
        #Then we remove this effect
        Player.EffectToRemove.append(ManaStackEffect)
    elif Spell.GCD and not (Spell.id == EnchantedRiposte.id or Spell.id == EnchantedZwerchhau.id or Spell.id == EnchantedRedoublement.id):
        #If not any of these, we loose mana stacks
        Player.ManaStack = 0
        Player.EffectToRemove.append(ManaStackEffect)

def RiposteCombo(Player, Spell):
    if Spell.id == EnchantedZwerchhau.id:
        Spell.Potency += 190
        Player.EffectToRemove.append(RiposteCombo)
        Player.EffectList.append(ZwerchhauCombo)
    elif Spell.id == Zwerchhau.id:
        Spell.Potency += 50
        Player.EffectToRemove.append(RiposteCombo)
        Player.EffectList.append(ZwerchhauCombo)

def ZwerchhauCombo(Player, Spell):
    if Spell.id == EnchantedRedoublement.id:
        Spell.Potency += 370
        Player.EffectToRemove.append(ZwerchhauCombo)
    elif Spell.id == Redoublement.id:
        Spell.Potency += 130
        Player.EffectToRemove.append(ZwerchhauCombo)

def DualCastEffect(Player, Spell):
    if Spell.CastTime != 0 and  Spell.GCD and Spell.id != -1:   #Want to make sure the spell will affect Dualcast, id != -1 is to make sure this is not WaitAbility
        if Player.DualCast:
            Spell.CastTime = 0 #Insta cast half of the spells, will be put by default for RedMage
            Player.DualCast = False #Remove Dualcast
        elif not (SwiftcastEffect in Player.EffectList): Player.DualCast = True    #Give Dual cast only if not insta from swift (Acc.'s case should be taken care off before that time since its insert(0))

def ManaficationEffect(Player, Spell):
    if Spell.GCD : #Only affect if GCD
        Player.ManaficationStack -= 1
        Spell.DPSBonus *= 1.05 #5% boost on magic damage

def AccelerationEffect(Player, Spell):
    if (Spell.id == Verthunder.id or Spell.id == Verareo.id or Spell.id == Impact.id):
        Spell.CastTime = 0 
        Player.EffectToRemove.append(AccelerationEffect)
        if Spell.id == Impact.id : Spell.Potency += 50 #Impact has high potency when used with Acceleration

        # Adding to Expected proc value

        match Spell.id:
            case Verthunder.id:
                Player.ExpectedVerfireProc += 1
            case Verareo.id:
                Player.ExpectedVerstoneProc += 1

#Check

def ManaficationCheck(Player, Enemy):
    if Player.ManaficationStack == 0: 
        Player.EffectList.remove(ManaficationEffect)#If no stack remove it
        Player.EffectToRemove.append(ManaficationCheck)

def EmboldenCheck(Player, Enemy):
    if Player.EmboldenTimer <= 0:
        Player.EmboldenTimer = 0
        Enemy.buffList.remove(EmboldenBuff) #Removing 5% dps increase
        Player.EffectToRemove.append(EmboldenCheck)

def AccelerationStackCheck(Player, Enemy):
    if Player.AccelerationCD <= 0:
        Player.AccelerationStack += 1

        if Player.AccelerationStack == 2:
            Player.EffectToRemove.append(AccelerationStackCheck)
        else:
            Player.AccelerationCD = 55

def EngagementStackCheck(Player, Enemy):
    if Player.EngagementCD <= 0:
        Player.EngagementStack += 1
        if Player.EngagementStack == 2:
            Player.EffectToRemove.append(EngagementStackCheck)
        else:
            Player.EngagementCD = 35

def CorpsStackCheck(Player, Enemy):
    if Player.CorpsCD <= 0:
        Player.CorpsStack += 1

        if Player.CorpsStack == 2:
            Player.EffectToRemove.append(CorpsStackCheck)
        else:
            Player.CorpsCD = 35


#GCD    

Jolt = RedmageSpell(7524, True, 2, 2.5,320, 200, ApplyJolt, [ManaRequirement], 0, 0, type = 1)
Verfire = RedmageSpell(7510, True, 2, 2.5, 340, 200, ApplyVerfire, [ManaRequirement], 0, 0, type = 1)
Verstone = RedmageSpell(7511, True, 2, 2.5, 340, 200, ApplyVerstone, [ManaRequirement], 0, 0, type = 1)
Verthunder = RedmageSpell(25855, True, 5, 2.5, 380, 300, ApplyVerthunder, [ManaRequirement], 0, 0, type = 1)
Verareo = RedmageSpell(25856, True, 5, 2.5, 380, 300, ApplyVerareao, [ManaRequirement], 0, 0, type = 1)
#AoEs
Impact = RedmageSpell(16526, True, 5, 2.5, 210, 400, ApplyImpact, [ManaRequirement], 0, 0, type = 1)
#Combo actions
#NonEnchanted
Riposte = RedmageSpell(7504, True, 0, 2.5, 130, 0, ApplyRiposte, [], 0, 0, type = 2)
Zwerchhau = RedmageSpell(7512, True, 0, 2.5, 100, 0, empty, [], 0, 0, type = 2)
Redoublement = RedmageSpell(7516, True, 0, 2.5, 100, 0, empty, [], 0, 0, type = 2)
Reprise = RedmageSpell(16529, True, 0, 2.5, 100, 0, empty, [], 0, 0, type = 2)
#Enchanted
EnchantedRiposte = RedmageSpell(7527, True, 0, 1.5, 280, 0, ApplyEnchantedRiposte, [RDMManaRequirement], 20,20, type = 2)
EnchantedZwerchhau = RedmageSpell(7528, True, 0, 1.5, 150, 0, ApplyZwerchhau, [RDMManaRequirement], 15, 15, type = 2)
EnchantedRedoublement = RedmageSpell(7529, True, 0, 2.2, 130, 0, ApplyRedoublement, [RDMManaRequirement], 15, 15, type = 2)
EnchantedReprise = RedmageSpell(16528, True, 0, 2.5, 340, 0, empty, [RDMManaRequirement], 5, 5, type = 2)
Verholy = RedmageSpell(7526, True, 0, 2.5, 600, 400, ApplyVerholy, [ManaRequirement, VerholyRequirement], 0, 0, type = 1)
Verflare = RedmageSpell(7525, True, 0, 2.5, 600, 400, ApplyVerflare, [ManaRequirement, VerholyRequirement], 0, 0, type = 1) #Same Requirement as Verholy, just need 3 Mana stacks
Scorch = RedmageSpell(16530, True, 0, 2.5, 680, 400, ApplyScorch, [ManaRequirement, ScorchRequirement], 0, 0, type = 1)
Resolution = RedmageSpell(25858, True, 0, 2.5, 750, 400, ApplyResolution, [ManaRequirement, ResolutionRequirement], 0, 0, type = 1)
#AOE Melee Action
Moulinet = RedmageSpell(7513, True, 0, 1.5, 130, 0, ApplyMoulinet, [RDMManaRequirement], 20, 20, type = 2)
EnchantedMoulinet = RedmageSpell(7530, True, 0, 1.5, 130, 0, ApplyEnchantedMoulinet, [RDMManaRequirement], 20, 20, type = 2)
#For now combo action cannot be cancelled by doing something else
Manafication = RedmageSpell(7521, False, 0, Lock, 0, 0, ApplyManafication, [ManaficationRequirement], 0, 0)
Embolden = RedmageSpell(7520, False, 0, Lock, 0, 0, ApplyEmbolden, [EmboldenRequirement], 0, 0)
Acceleration = RedmageSpell(7518, False, 0, Lock, 0, 0, ApplyAcceleration, [AccelerationRequirement], 0, 0)
Fleche = RedmageSpell(7517, False, 0, Lock, 460, 0, ApplyFleche, [FlecheRequirement], 0, 0)
Contre = RedmageSpell(7519, False, 0, Lock, 380, 0, ApplyContre, [ContreRequirement], 0, 0)
Engagement = RedmageSpell(16527, False, 0, Lock, 180, 0, ApplyEngagement, [EngagementRequirement], 0, 0)
Corps = RedmageSpell(7506, False, 0, Lock, 130, 0, ApplyCorps, [CorpsRequirement], 0, 0)

#Other GCD/oGCD with no DPS goal
MagickBarrier = RedmageSpell(25857, False, 0, 0, 0, 0, ApplyMagickBarrier, [MagickBarrierRequirement], 0, 0)
Verraise = RedmageSpell(7523, True,10, 2.5, 0,2400, empty, [RDMManaRequirement], 0, 0, type = 1)
Vercure = RedmageSpell(7514, True, 2, 2.5, 0, 500, empty, [RDMManaRequirement], 0, 0, type = 1)



#buff
EmboldenBuff = buff(1.05,name="Emb")

#ActionList
RedMageAbility = {
7530 : EnchantedMoulinet,
7504 : Riposte, 
7506 : Corps, 
7507 : Verareo, 
7510 : Verfire, 
7511 : Verstone, 
7512 : Zwerchhau, 
7513 : Moulinet, 
7514 : Vercure, 
7515 : Engagement,
7516 : Redoublement, 
7517 : Fleche, 
7518 : Acceleration, 
7519 : Contre,
7520 : Embolden, 
7521 : Manafication , 
7523 : Verraise, 
7524 : Jolt, 
25855 : Verthunder, 
25856 : Verareo, 
16526 : Impact, 
16527 : Engagement, 
16529 : Reprise,  
25857 : MagickBarrier, 
7527 : EnchantedRiposte, 
7528 : EnchantedZwerchhau,
7529 : EnchantedRedoublement,  
7530 : EnchantedMoulinet, 
16528 : EnchantedReprise,
25858 : Resolution, 
7526 : Verholy, 
7525 : Verflare, 
16530 : Scorch
}


"""
This code was written by Apollo (apollo.van.waddleburg on discord) and adapted to fit the needs of the sim.
"""

"""Fetches log data from a fflogs URL."""
import logging
import os
from typing import Dict, List, Tuple, Iterator, Optional

import requests
import dataclasses
import python_graphql_client
import pandas as pd

from ffxivcalc.Request import custom_columns
from ffxivcalc.helperCode.exceptions import invalidFFLogsFightId, invalidFFLogsQuery
from ffxivcalc.Request.ffxivcalcViewTranslator import ffxiv_sim_view
LOG = logging.getLogger(__name__)

def getSingleFightData(clientId : str, clientSecret : str, code : str, id : str, showProgress : bool = False, max_time : float = 0) -> Dict:
    """Gets a fight data from fflog and returns the data corresponding to that fight in a format ffxivcalc can use.
       Errors will be diplayed.
       
    clientId : str - Your client Id from fflogs. If left empty doesn't write to environ.
    clientSecret : str - Your client secret from fflogs. If left empty doesn't write to environ.
    code : str - Code of the fights record.
    id : str - Id of the fight.
    showProgress : bool = False - If true a progress bar will be shown.
    max_time : float = 0 - Max time stamp to take from fflogs. If left as default (0) it does not limit time stamp.
    """
    if len(clientId) != 0 : os.environ["FFLOGS_CLIENT_ID"] = clientId
    if len(clientSecret) != 0 : os.environ["FFLOGS_CLIENT_SECRET"] = clientSecret
    client = FFLogClientV2()

    for rawFight in client.stream_fight_events(code, fight_id=id, show_progress=showProgress):
        df = get_fflog_events_dataframe(rawFight[1])
        view = ffxiv_sim_view(df, max_time=max_time)
        for fight in view:
            data = fight[1]
            _aura_prepull_logic(data)
            return data
        
def _aura_prepull_logic(fight):
        # Aura prepull logic.
    foundLeftEye = False
    foundRightEye = False
    rightEyeId = -1
    leftEyeTargetId = -1

    foundDancePartner = False
    foundClosedPosition = False
    dancerId = -1
    partnerId = -1

    for player in fight["data"]["PlayerList"]:
        for aura in player["Auras"]:
            if aura == "SharpCast":
                player["actionList"].insert(0,{"actionName" : "Sharpcast"})
            elif aura == "Ley Lines":
                player["actionList"].insert(0,{"actionName" : "LeyLines"})
            elif aura == "Triplecast":
                player["actionList"].insert(0,{"actionName" : "Triplecast"})
            elif aura == "Medicated":
                player["actionList"].insert(0,{"actionName" : "Potion"})
            elif aura == "Inner Release":
                player["actionList"].insert(0,{"actionName" : "InnerRelease"})
            elif aura == "Fight or Flight":
                player["actionList"].insert(0,{"actionName" : "FightOrFlight"})
            elif aura == "Delirium":
                player["actionList"].insert(0,{"actionName" : "Delirium"})
            elif aura == "Blood Weapon":
                player["actionList"].insert(0,{"actionName" : "BloodWeapon"})
            elif aura == "Blackest Night" and player["JobName"] == "DarkKnight":
                player["actionList"].insert(0,{"actionName" : "TheBlackestNight", "targetID" : player["playerID"]})
                # TODO Add code that makes the darkknight cast it. Now only casts if self cast from DRK
            elif aura == "No Mercy":
                player["actionList"].insert(0,{"actionName" : "NoMercy"})
            elif aura == "Arcane Circle":
                if player["JobName"] == "Reaper":
                    player["actionList"].insert(0,{"actionName" : "ArcaneCircle"})
            elif aura == "Soulsow":
                player["actionList"].insert(0,{"actionName" : "Soulsow"})
            elif aura == "Riddle of Fire":
                player["actionList"].insert(0,{"actionName" : "RiddleOfFire"})
            elif aura == "Riddle of Wind":
                player["actionList"].insert(0,{"actionName" : "RiddleOfWind"})
            elif aura == "Brotherhood":
                player["actionList"].insert(0,{"actionName" : "Brotherhood"})
            elif aura == "Life Surge":
                player["actionList"].insert(0,{"actionName" : "LifeSurge"})
            elif aura == "Lance Charge":
                player["actionList"].insert(0,{"actionName" : "LanceCharge"})
            elif aura == "Battle Litany" and player["JobName"] == "Dragoon":
                player["actionList"].insert(0,{"actionName" : "BattleLitany"})
            elif aura == "Right Eye":
                if not foundLeftEye:
                    foundRightEye = True
                    rightEyeId = player["playerID"]
                else:
                    player["actionList"].insert(0,{"actionName" : "DragonSight", "targetID" : leftEyeTargetId})
            elif aura == "Left Eye":
                if not foundRightEye:
                    foundLeftEye = True
                    leftEyeTargetId = player["playerID"]
                else:
                    # Looking for player
                    for emitPlayer in fight["data"]["PlayerList"]:
                        if emitPlayer["playerID"] == rightEyeId:
                            emitPlayer["actionList"].insert(0,{"actionName" : "DragonSight", "targetID" : player["playerID"]})
            elif aura == "Kassatsu":
                player["actionList"].insert(0,{"actionName" : "Kassatsu"})
            elif aura == "Meikyo Shisui":
                player["actionList"].insert(0,{"actionName" : "Meikyo"})
            elif aura == "Battle Voice" and player["JobName"] == "Bard":
                player["actionList"].insert(0,{"actionName" : "BattleVoice"})
            elif aura == "Acceleration":
                player["actionList"].insert(0,{"actionName" : "Acceleration"})
            elif aura == "Embolden" and player["JobName"] == "RedMage":
                player["actionList"].insert(0,{"actionName" : "Embolden"})
            elif aura == "Searing Light" and player["JobName"] == "Summoner":
                player["actionList"].insert(0,{"actionName" : "SearingLight"})
            elif aura == "Presence of Mind":
                player["actionList"].insert(0,{"actionName" : "PresenceOfMind"})
            elif 'Drawn' in aura:
                player["actionList"].insert(0,{"actionName" : "Draw"})
            elif aura == "Earthly Dominance":
                player["actionList"].insert(0,{"actionName" : "EarthlyStar"})
            elif aura == "Defiance":
                player["actionList"].insert(0,{"actionName" : "Defiance"})
            elif aura == "Royal Guard":
                player["actionList"].insert(0,{"actionName" : "RoyalGuard"})
            elif aura == "Grit":
                player["actionList"].insert(0,{"actionName" : "Grit"})
            elif aura == "Iron Will":
                player["actionList"].insert(0,{"actionName" : "IronWill"})
            elif aura == "Mudra": # This has nothing to do with Huton but if we see that aura we will give it.
                player["actionList"].insert(0,{"actionName" : "Huton"})
            elif aura == "Reassembled":
                player["actionList"].insert(0,{"actionName" : "Reassemble"})
            elif aura == "Eukrasia":
                player["actionList"].insert(0,{"actionName" : "Eukrasia"})
            elif aura == "Standard Step":
                # Assume 2 steps
                player["actionList"].insert(0,{"actionName" : "Entrechat"})
                player["actionList"].insert(0,{"actionName" : "Jete"})
                player["actionList"].insert(0,{"actionName" : "StandardStep"})
            elif aura == "Technical Step":
                # Assume 4 steps
                player["actionList"].insert(0,{"actionName" : "Emboite"})
                player["actionList"].insert(0,{"actionName" : "Pirouette"})
                player["actionList"].insert(0,{"actionName" : "Entrechat"})
                player["actionList"].insert(0,{"actionName" : "Jete"})
                player["actionList"].insert(0,{"actionName" : "TechnicalStep"})
            elif aura == "Closed Position":
                if not foundDancePartner:
                    foundClosedPosition = True
                    dancerId = player["playerID"]
                else:
                    player["actionList"].insert(0,{"actionName" : "ClosedPosition", "targetID" : partnerId})
            elif aura == "Dance Partner":
                if not foundClosedPosition:
                    foundDancePartner = True
                    partnerId = player["playerID"]
                else:
                    # Looking for player
                    for emitPlayer in fight["data"]["PlayerList"]:
                        if emitPlayer["playerID"] == dancerId:
                            emitPlayer["actionList"].insert(0,{"actionName" : "ClosedPosition", "targetID" : player["playerID"]})
        player["Auras"] = [] # Wiping auras.


def get_auth_token() -> str:
    """Fetches an auth token via https://www.fflogs.com/oauth/token.

    See https://www.fflogs.com/api/docs/

    And https://www.fflogs.com/v2-api-docs/ff/report.doc.html

    Requires FFLOGS_CLIENT_ID and FFLOGS_CLIENT_SECRET set as environment variables.
    These can be obtained here: https://www.fflogs.com/api/clients/

    :returns: Auth token for FFLogs APIv2
    """
    LOG.debug(f"Fetching fflogs auth token")
    x = requests.post(
        url="https://www.fflogs.com/oauth/token",
        data={"grant_type": "client_credentials"},
        auth=(
            os.environ["FFLOGS_CLIENT_ID"],
            os.environ["FFLOGS_CLIENT_SECRET"],
        ),
    )

    x.raise_for_status()
    return x.json()["access_token"]

def get_fflog_events_dataframe(data: List) -> pd.DataFrame:
    """Converts fflog events into a dataframe with timestamp set."""
    event_data = pd.DataFrame(data)

    # Run lb-update hard sync to ensure damage taken instances are consistent
    potential_start_times = [
        event_data[event_data[custom_columns.TYPE] == custom_columns.DAMAGE][
            "timestamp"
        ].min()
    ]
    if event_data[custom_columns.TYPE].isin(["limitbreakupdate"]).any():
        potential_start_times.append(
            event_data[event_data[custom_columns.TYPE] == "limitbreakupdate"][
                "timestamp"
            ].min()
        )

    start_timestamp = min(potential_start_times)

    event_data.insert(
        0,
        custom_columns.FIGHT_TIME,
        ((event_data["timestamp"] - start_timestamp) / 1000.0).round(4),
    )
    del event_data["timestamp"]
    return event_data

@dataclasses.dataclass(frozen=True, repr=False)
class FFLogClientV2:
    """Base class for setting up an FFLogsClient."""

    client: python_graphql_client.GraphqlClient = dataclasses.field(
        default_factory=lambda: (
            python_graphql_client.GraphqlClient(
                endpoint="https://www.fflogs.com/api/v2/client",
                headers={"Authorization": f"Bearer {get_auth_token()}"},
            )
        )
    )

    # Caches outcomes from https://www.fflogs.com/v2-api-docs/ff/reportability.doc.html
    # Maps IDs to ability data
    # Structure:
    # {
    #     log_code: {
    #         ability_id (str): ability_data (dict)
    #         ...
    #     },
    #     ...
    # }
    ability_id_cache: Dict = dataclasses.field(default_factory=dict)

    # Caches outcomes from https://www.fflogs.com/v2-api-docs/ff/reportactor.doc.html
    # Maps ids to actor data
    actor_cache: Dict = dataclasses.field(default_factory=dict)

    def get_fights_enriched(self, code: str, fight_type: str) -> Dict:
        """Fetches the high level fights from a log. Contains more info than normal.

        Output contains info from both report and the fights.
        """
        LOG.debug(f"Fetching fflogs reportData:report:fights via {code}")
        data = self.client.execute(
            query="""query reportData($code: String!, $kills: KillType!){
                  reportData
                  {
                    report(code: $code){
                      endTime
                      startTime
                      guild{
                        name
                      }
                      fights(killType: $kills){
                        encounterID
                        startTime
                        endTime
                        name
                        kill
                        standardComposition
                        combatTime
                        id
                      }
                    }
                  }
                }
                """,
            variables={
                "code": code,
                "kills": fight_type,
            },
        )
        self._check_if_valid_query(data)
        return data["data"]["reportData"]["report"]

    def _check_if_valid_query(self,query : Dict):
        """Checks the validity of the returned query. If a query is invalid it contains the key 'message' with a message. An error
        will be raised an the message displayed to the user."""

        if 'errors' in query.keys():
            raise invalidFFLogsQuery(query['errors'][0]['message'])

    def _check_valid_fight_id(self,fightList : List[Dict], fight_id : str) -> bool:
        """Checks if the given id is within the retrieved fight list. Raises invalidFFLogsFightId if not valid."""

        if not fight_id in [
            str(fight['id']) for fight in fightList
        ]:
            raise invalidFFLogsFightId(fight_id)

    def get_fights(self, code: str,fight_type: str = "Kills") -> List[Dict]:
        """Fetches the high level fights from a log."""
        LOG.debug(f"Fetching fflogs reportData:report:fights via {code}")
        data = self.client.execute(
            query="""query reportData($code: String!, $kills: KillType!){
                  reportData
                  {
                    report(code: $code){
                      fights(killType: $kills){
                        id
                        encounterID
                        startTime
                        endTime
                        name
                        kill
                        standardComposition
                        combatTime
                      }
                    }
                  }
                }
                """,
            variables={
                "code": code,
                "kills": fight_type
            },
        )
        self._check_if_valid_query(data)

        return data["data"]["reportData"]["report"]["fights"]

    def get_abilities(self, code: str) -> List[Dict]:
        """Fetches the high level fights from a log."""
        LOG.debug(f"Fetching fflogs reportData:report:fights via {code}")
        data = self.client.execute(
            query="""query reportData($code: String!){
                  reportData
                  {
                    report(code: $code){
                      masterData{
                        abilities{
                          gameID
                          name
                          type
                        }
                      }
                    }
                  }
                }
                """,
            variables={
                "code": code,
            },
        )
        self._check_if_valid_query(data)

        return data["data"]["reportData"]["report"]["masterData"]["abilities"]

    def stream_fight_events(
        self, code: str, fight_type: str = "Kills", fight_id : str = "", show_progress : bool = False
    ) -> Iterator:
        """Streams fight event data

        Structure:
            yield fight_name, [{events...}]
        """
        self._populate_caches(code=code)

        fightList = self.get_fights(code=code,fight_type=fight_type)

        if len(fight_id) != 0: self._check_valid_fight_id(fightList, fight_id)

        for fight in fightList:
            if fight_id != "" and str(fight["id"]) != fight_id : continue
            yield (
                fight["name"],
                self._process_fight_data(
                    data=self.get_raw_events_from_fight(
                        code=code,
                        fight=fight,
                        fight_type=fight_type,
                        show_progress=show_progress
                    ),
                    code=code,
                ),
            )

    def _process_fight_data(self, data: Iterator, code: str) -> List:
        """Applies actor and ability name replacement."""
        return [
            self._process_raw_event(data=raw_data_event, code=code)
            for raw_data_event in data
        ]

    def _process_raw_event(self, data: Dict, code: str) -> Dict:
        
        if "extraAbilityGameID" in data: # Sometime early potions appear as a weird ID with the correct value in'extraAbilityGameID'. So we copy this value if it happens here.
            data["abilityGameID"] = data["extraAbilityGameID"]

        if "abilityGameID" in data:
            data[custom_columns.ABILITY_NAME] = self._get_ability(
                data["abilityGameID"], code=code
            )

        if "sourceID" in data:
            data[custom_columns.SOURCE_NAME] = self._get_actor_name(
                data["sourceID"], code=code
            )
            source_actor = self._get_actor(data["sourceID"], code=code)
            if source_actor:
                data[custom_columns.SOURCE_TYPE] = source_actor["type"]
                data[custom_columns.SOURCE_SUBTYPE] = source_actor["subType"]
                if source_actor["petOwner"]:
                    data[
                        custom_columns.SOURCE_PET_ACTOR
                    ] = self._get_actor_name(
                        source_actor["petOwner"], code=code
                    )

        if "targetID" in data:
            data[custom_columns.TARGET_NAME] = self._get_actor_name(
                data["targetID"], code=code
            )
            target_actor = self._get_actor(data["targetID"], code=code)
            if target_actor:
                data[custom_columns.TARGET_TYPE] = target_actor["type"]
                data[custom_columns.TARGET_SUBTYPE] = target_actor["subType"]
                if target_actor["petOwner"]:
                    data[
                        custom_columns.TARGET_PET_ACTOR
                    ] = self._get_actor_name(
                        target_actor["petOwner"], code=code
                    )

        return data

    def _get_ability(self, ability_id: str, code: str) -> str:
        if code not in self.ability_id_cache:
            self._populate_caches(code)

        return (
            "UNKNOWN"
            if ability_id not in self.ability_id_cache[code]
            else (self.ability_id_cache[code][ability_id]["name"])
        )

    def _get_actor(self, actor_id: str, code: str) -> Optional[Dict]:
        """Gets the full actor data."""
        if code not in self.actor_cache:
            self._populate_caches(code)

        return self.actor_cache[code].get(actor_id)

    def _get_actor_name(self, actor_id: str, code: str) -> Optional[str]:
        """Gets the full actor name. Adjusts pet names to include owner."""
        actor: Optional[Dict] = self._get_actor(actor_id=actor_id, code=code)
        if actor is None:
            return None

        if not actor["petOwner"]:
            return actor["name"]

        pet_owner: Optional[Dict] = self._get_actor(
            actor_id=actor["petOwner"], code=code
        )["name"]

        return f"{actor['name']} ({pet_owner})"

    def _populate_caches(self, code: str) -> None:
        LOG.debug(f"Fetching fflogs reportData:report:masterData via {code}")
        log_data = self.client.execute(
            query="""query reportData($code: String!){
                          reportData
                          {
                            report(code: $code){
                              masterData{
                                abilities {
                                  gameID
                                  name
                                  type
                                }
                                actors {
                                  id
                                  name
                                  petOwner
                                  subType
                                  type
                                }
                              }
                            }
                          }
                        }
                        """,
            variables={"code": code},
        )
        self._check_if_valid_query(log_data)

        log_data = log_data["data"]["reportData"]["report"]["masterData"]

        self.ability_id_cache[code] = {
            ability["gameID"]: ability for ability in log_data["abilities"]
        }

        self.actor_cache[code] = {
            actor["id"]: actor for actor in log_data["actors"]
        }

        LOG.debug(f"Stored actor and ability ids.")

        self._update_id_data()

        return

    def _update_id_data(self):
        """Updates the current database csv of ability ids."""
        # flatten the dictionary to remove codes
        updates = {}
        for ids_to_abilities in self.ability_id_cache.values():
            updates = {
                **updates,
                **{
                    ability["gameID"]: ability
                    for ability in ids_to_abilities.values()
                },
            }
        #game_data.update_ability_ids_to_names(updates)

    def get_raw_events_from_fight(
        self, code: str, fight: Dict, fight_type: str = "Kills", show_progress : bool = False
    ) -> Iterator:
        """Gets the raw events from fight."""
        next_timestamp = start_time = fight["startTime"]
        end_time = fight["endTime"]
        LOG.debug(
            f"Fetching fflogs reportData:report:events via {code} [{start_time}, {end_time}]"
        )

        # Pretty animation

        percent_complete = (
            "{:5.2f}%".format(
                (float(next_timestamp) - float(start_time))
                / (float(end_time) - float(start_time))
                * 100.0
            )
            if (float(end_time) > float(start_time))
            else "100%"
        )
        animation: str = f"Loading {fight['name']} ({percent_complete})"
        while next_timestamp is not None and int(next_timestamp) < int(
            end_time
        ):
            # Pretty animation
            if show_progress: print(animation, end="\r")
            page_data, next_timestamp = self._get_events_from_fight(
                code=code,
                start_time=next_timestamp,
                end_time=end_time,
                fight_type=fight_type,
            )

            for event_data in page_data:
                yield event_data

            percent_complete = "{:5.2f}%".format(
                (float(next_timestamp or end_time) - float(start_time))
                / (float(end_time) - float(start_time))
                * 100.0
            )
            animation: str = f"Loading {fight['name']} ({percent_complete})"
        if show_progress: print(f"{animation}: DONE")

    def _get_events_from_fight(
        self,
        code: str,
        start_time: str,
        end_time: str,
        fight_type: str = "Kills",
    ) -> Tuple:
        """Inner function for the timestamp management."""
        raw_data = self.client.execute(
            query="""query reportData($code: String!, $startTime: Float!, $endTime: Float!, $kills: KillType!, $limit: Int!){
                          reportData
                          {
                            report(code: $code){
                              events(killType: $kills, startTime: $startTime, endTime: $endTime, limit: $limit){
                                data
                                nextPageTimestamp
                              }
                            }
                          }
                        }
                        """,
            variables={
                "code": code,
                "startTime": start_time,
                "endTime": end_time,
                "kills": fight_type,
                "limit": 10000,
            },
        )

        self._check_if_valid_query(raw_data)
        raw_data = raw_data["data"]["reportData"]["report"]["events"]

        return raw_data["data"], raw_data["nextPageTimestamp"]